// $Id: Sequence.cpp,v 1.11 2005/10/10 22:37:30 mkkuhner Exp $

/* 
 Copyright 2002 Patrick Colacurcio, Peter Beerli, Mary Kuhner, 
                Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "Converter_Sequence.h"
#include "Converter_DataSourceException.h"
#include "random.h"

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

using namespace std;

//____________________________________________________

Sequence::Sequence (const string &src)
{
    m_sequence = AsStringVec(src);
    trimSequence();
    string sequence = asString();
    validate(sequence);
    setDataType ( "DNA" );  // Default Data Type
    Random random; // uses system clock
    setName ( random.Name() );
}				/* Sequence::Sequence */

Sequence::Sequence (const string &src,
		    const string &dataType,
                    const string &name)
{
      if (dataType == lamarcstrings::DNA || dataType == lamarcstrings::SNP)
        {
          m_sequence = AsStringVec(src);
          trimSequence();
        }
      else if (dataType == lamarcstrings::MICROSAT || dataType == lamarcstrings::KALLELE)
        {
          m_sequence.push_back(src);
        }

      if (dataType == lamarcstrings::DNA)
	{
          string sequence = asString();
	  validate(sequence);
	}

      setDataType ( dataType );
      setName ( name );
}				/* Sequence::Sequence */

Sequence::Sequence (const vector<string> &src,
		    const string &dataType,
                    const string &name)
{
      m_sequence = src;
      if (dataType == lamarcstrings::DNA || dataType == lamarcstrings::SNP ) trimSequence();

      if (dataType == lamarcstrings::DNA)
	{
          string sequence = asString();
	  validate(sequence);
	}

      setDataType ( dataType );
      setName ( name );
}				/* Sequence::Sequence */

//____________________________________________________

Sequence::~Sequence ()
{
}				/* Sequence::~Sequence */

//____________________________________________________

// If only the standard allowed for a string::string(char&) ctor,
// then we'd just use 
//   vector<string>::assign(src.begin(),src.end()) 
// in place of this whole messy function!
vector<string> Sequence::AsStringVec(const string& src) const
{
  vector<string> sequence;
  string::const_iterator marker;
  for(marker = src.begin(); marker != src.end(); ++marker)
    sequence.push_back(string(1L,*marker));

  return sequence;

} /* AsStringVec */

//____________________________________________________

string
Sequence::asString() const
{
  string sequence;
  vector<string>::const_iterator seq;
  for(seq = m_sequence.begin(); seq != m_sequence.end(); ++seq)
    sequence += *seq;
  return sequence;
}

// ____________________________________________________

long
Sequence::getSequenceLength() const
{
  long length = m_sequence.size();
  return length;
}

void
Sequence::setDataType( const string &dataType)
{
  m_dataType = dataType;
}

void
Sequence::validate(const string &src) const
{
  int n = src.length();
  int illegalPosition;

  illegalPosition = src.find_first_not_of ("agctumrwsykvhdbnox?-AGCTUMRWSYKVHDBNOX");

  if ((illegalPosition >= 0) && (illegalPosition < n))
      {
	{
	  string err 
	    ("Invalid character in sequence: ");
	  err += src[illegalPosition];
	  err += ";";

	  throw InvalidNucleotideError(err);
	}

    }
}

void
Sequence::trimSequence()
{
  string sequence = asString();
  // Gets rid of the whitespace
  size_t beg = sequence.find_first_not_of(" \n\t");
  size_t end = sequence.find_last_not_of(" \n\t");

  // Mary fix:  off by one error (STL uses half-open intervals) 2/20/02
  sequence = sequence.substr(beg, end - beg + 1);

  m_sequence = AsStringVec(sequence);
}

string
Sequence::getXML(unsigned int numTabs) const
{
  string sequenceXML;

  addTabs(numTabs, sequenceXML);  
  sequenceXML += "<sample name=\"" + m_name + "\">\n";
  
  ++numTabs;
  addTabs(numTabs, sequenceXML);  
  sequenceXML += "<datablock type=\"" + m_dataType + "\">\n";

  ++numTabs;
  addTabs(numTabs, sequenceXML);  
  string sequence;
  if (m_dataType == lamarcstrings::DNA || m_dataType == lamarcstrings::SNP)
    sequence = asString();
  else
    {
      vector<string>::const_iterator mseq;
      for(mseq = m_sequence.begin(); mseq != m_sequence.end(); ++mseq)
        {
          sequence += *mseq + " ";
        }
    }
  sequenceXML += sequence;
  sequenceXML += "\n";

  --numTabs;
  addTabs(numTabs, sequenceXML);  
  sequenceXML += "</datablock>\n";

  --numTabs;
  addTabs(numTabs, sequenceXML);  
  sequenceXML += "</sample>\n";

  return sequenceXML;
}
