// $Id: SpaceConverter.cpp,v 1.10 2005/01/20 00:14:44 ewalkup Exp $

/* 
 Copyright 2002 Patrick Colacurcio, Peter Beerli, Mary Kuhner, 
                Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "Converter_SpaceConverter.h"
#include "Converter_DataSourceException.h"
#include "Converter_LamarcDS.h"
#include "stringx.h"

#include <vector>
#include <iostream>
#include <sstream>

//----------------------------------------------------------------------
SpaceConverter::SpaceConverter(LamarcDS& lamarc) :
ConverterIf(), m_lamarc(lamarc)
{
} /* SpaceConverter::ctor */

//----------------------------------------------------------------------

SpaceConverter::~SpaceConverter()
{
} /* SpaceConverter::dtor */

//----------------------------------------------------------------------

void SpaceConverter::ValidateSpaceInfo(const SpaceMap& spaces,
                                       const string& fname) const
{
  // now make sure that every region that is addressed in the space info
  // exists in the data under the same name
  SpaceMap::const_iterator spaceit;
  for(spaceit = spaces.begin(); spaceit != spaces.end(); ++spaceit)
    {
      if (!m_lamarc.doesRegionNameExist(spaceit->first))
        {
          string errormsg = "Region name " + spaceit->first;
          errormsg += " was mentioned in file " + fname;
          errormsg += "\nbut no corresponding region name was found in the data.\n";
          errormsg += "Make sure that region names are used";
          errormsg += " consistently in all parts of data entry.";
          throw FileFormatError(errormsg);
        }
    }

} /* SpaceConverter::ValidateSpaceInfo */

//----------------------------------------------------------------------

SpaceMap SpaceConverter::ReadSpacingInfo(const string& filename) const
{
  SpaceMap spacings;

  ifstream spacefile(filename.c_str(),ios::in);

  if (!spacefile)
    {
      string errormsg = "I could not find file, " + filename;
      throw FileFormatError(errormsg);
    }

  string line, somestring;
  if (!getLine(spacefile,line))
    {
      string errormsg = "Your file, " + filename + ", appears";
      errormsg += " to be empty.";
      throw FileFormatError (errormsg);
    }

  bool notdone = true;
  bool firsttime  = true;
  long linenumber = 1;
  string regionname;
  vector<long> positions;

  // now parse the file line by line
  while(notdone)
    {
      istringstream linestream(line);

      skipWhiteSpace(linestream);
      somestring.erase();

      // process the first token in the line.
      if (getName(linestream,somestring))
        {
          if (!firsttime)  // insert previous finished region
            spacings.insert(make_pair(regionname,positions));
          firsttime = false;
          regionname = somestring;
          positions.clear();
        }
      else
        {
          if (IsInteger(somestring))
            {
              long position = FlagCheck(somestring,
                                        string("position number"));
              positions.push_back(position);
            }
          else
            {
               string errormsg = "An illegal region name was found";
               errormsg += " on line " + ToString(linenumber);
               errormsg += " of file " + filename +"\nTo be legal";
               errormsg += " a region name must have at least one";
               errormsg += " alphabetic character in it.";
               throw FileFormatError (errormsg);
            }
        }

      // process the rest of the line, which must be numbers
      skipWhiteSpace(linestream);
      while (linestream.good())
        {
          somestring.erase();
          if (getNumber(linestream,somestring))
            {
              long position = FlagCheck(somestring,
                                        string("position number"));
              positions.push_back(position);
            }
          else
            {
              string errormsg = "The non-number " + somestring + " was";
              errormsg += " encountered while reading region " + regionname;
              errormsg += " in file " + filename + " on line ";
              errormsg += ToString(linenumber);
              throw FileFormatError (errormsg);
            }
          skipWhiteSpace(linestream);
        }

      line.erase();
      notdone = getLine(spacefile,line);
      ++linenumber;
    }
  // add the last region to the map
  spacings.insert(make_pair(regionname,positions));

#ifndef GUI
  ValidateSpaceInfo(spacings,filename);
#endif

  return spacings;

} /* SpaceConverter::ReadSpacingInfo */

//----------------------------------------------------------------------

void SpaceConverter::addConvertedLamarcDS(LamarcDS& lamarc)
{
  lamarc.mergeTo(m_lamarc);
} /* SpaceConverter::addConvertedLamarcDS */

//----------------------------------------------------------------------
