//$Id: range.cpp,v 1.19 2007/10/08 17:20:05 lpsmith Exp $
/* 
   Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

   This software is distributed free of charge for non-commercial use
   and is copyrighted.  Of course, we do not guarantee that the software
   works, and are not responsible for any damage you may cause or have.
*/

#include "range.h"
#include "stringx.h"
#include "errhandling.h"
#include "registry.h" //for random.bool

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

using namespace std;

//_______________________________________________________________
//_______________________________________________________________

Range::Range(long nsites, const vector<long>& alwaysact)
  : m_oldsites(),
    m_activesites(),
    m_newactivelinks(),
    m_oldfcsites(),
    m_fcsites(),
    m_recsite(FLAGLONG),
    m_newlinks(0L),
    m_alwaysactive(),
    m_side(unset_side),
    m_oldsitesset(false),
    m_oldfcset(false),
    m_totalsites(nsites)
{
  rangepair allsites(0L, nsites);
  m_activesites.clear();
  m_activesites.insert(allsites);
  // set up the local-partition-force locations
  for (unsigned long i = 0; i < alwaysact.size(); ++i) {
    pair<long,long> p = make_pair(alwaysact[i],alwaysact[i]+1);
    m_alwaysactive.insert(p);
  }

} /* Range::Range(long) */

//_______________________________________________________________

// Span returns the number of possible links in the range sent to it.
long Range::Span(const rangeset& sites) const
{
  if (sites.empty()) return 0L;

  long start = sites.begin()->first;
  long end = (--sites.end())->second;

  return end - start - 1;

} /* Range::Span */

//_______________________________________________________________

void Range::SetOldSitesIfNeeded()
{
  if (!m_oldsitesset) {
    m_oldsites = m_activesites;
    m_oldsitesset = true;
  }
  if (!m_oldfcset) {
    m_oldfcsites = m_fcsites;
    m_oldfcset = true;
  }
}

//_______________________________________________________
// At the end of this function, Range::m_newactivelinks
// should contain the sites that span newly active links,
// and Range::m_newlinks should contain a count of those links.

void Range::UpdateNewLinks()
{
  assert(m_oldsitesset);
  assert(m_oldfcset);
  
  m_newactivelinks.clear();
  m_newlinks = 0;

  //Any sites that have reached final coalescence should not be
  // considered (and were not considered in previous iterations).
  rangeset allactivesites = RemoveRangeFromRange(m_fcsites, m_activesites);
  rangeset alloldactivesites = RemoveRangeFromRange(m_oldfcsites, m_oldsites);

  //However, the alwaysactive sites always affect recombination eligibility,
  // regardless of FC.
  allactivesites = Union(allactivesites, m_alwaysactive);
  alloldactivesites = Union(alloldactivesites, m_alwaysactive);

  // If there are no active sites, there are no newly active links
  if(allactivesites.empty()) return;

  // Calculate the sites spanning the old active links and
  // the sites spanning the current active links

  long newstart = allactivesites.begin()->first;
  long newend = (--allactivesites.end())->second;
  if (newstart == newend-1) {
    return;
  }
  rangeset deadlinks;
  if (!alloldactivesites.empty()) {
    long oldstart = alloldactivesites.begin()->first;
    long oldend = (--alloldactivesites.end())->second;

    // if all links used to be active, none can be newly active, OR
    // if there is only 1 active site, there are no active links
    if ((oldstart == 0 && oldend == m_totalsites) || (newstart == newend-1))
      return;

    // Find the sites spanning the old *inactive* links
    if (oldstart > 0) deadlinks.insert(rangepair(0,oldstart));
    if (oldend < m_totalsites) deadlinks.insert(rangepair(oldend-1,m_totalsites-1));

    assert(!deadlinks.empty());
  }
  else {
    //If there used to be no active sites, all now-active sites newly-active
    deadlinks.insert(make_pair(0, m_totalsites-1));
  }
  rangeset livelinks;
  livelinks.insert(rangepair(newstart,newend-1));

  // Calculate the intersection of old dead links and current
  // live links; these are newly active links
  m_newactivelinks = Intersection(deadlinks, livelinks);

  // Compute the number of links involved
  rangeset::iterator i = m_newactivelinks.begin();
  for ( ; i != m_newactivelinks.end(); ++i) {
    m_newlinks += i->second - i->first;
  }

} /* Range::UpdateNewLinks */

//_________________________________________________

void Range::SetActiveFCWith(const rangeset& fc)
{
  //LS NOTE:  If things get confusing, the following line can be used instead
  // of the one following, so that any given range object will only know about
  // sites that have reached FC for sites that are actually currently active
  // on their corresponding branch.  This means that for any given site under
  // FC, exactly one branch's range object at a given time in the tree will
  // know that it is under FC.
  //
  // The advantage of this is that it keeps things a bit clearer when debugging.
  // The disadvantage is that it's much slower.

  //m_fcsites = Intersection(fc, m_activesites); //Slow, clear
  m_fcsites = fc;  //Fast, a bit more obscure
}

//_________________________________________________

void Range::ClearOldSites()
{
  m_oldsites.clear();
  m_oldsitesset = false;
  m_oldfcsites.clear();
  m_oldfcset = false;

} /* Range::ClearOldsites */

//_________________________________________________

void Range::ClearNewactiveLinks()
{
  m_newactivelinks.clear();
  m_newlinks = 0L;

} /* Range::ClearNewactivesites */

//_________________________________________________

bool Range::OldSitesAreLeft(long recsite)
{
  //This routine is used to tell whether a branch that has a new recombination
  // in it should keep the left half or right half in the tree.  We used to
  // decide randomly, which was legal, but which would decrease the acceptance
  // rate, due to essentially breaking this new branch instead of just
  // the initial chosen one.
  assert (m_oldsitesset);
  assert (m_oldsites.size() > 0);
  rangeset oldactive = RemoveRangeFromRange(m_oldfcsites, m_oldsites);
  //We don't care about m_alwaysactive since it will be active in both the
  // old branch and the new branch.
  if (oldactive.empty()) {
    //There were no un-supressed sites on the old branch, so it doesn't matter
    // which we pick this time.
    return registry.GetRandom().Bool();
  }
  return ((*oldactive.begin()).first <= recsite);
}

//_________________________________________________
//_________________________________________________

//The following routines only involve activesites and not alwaysactivesites.

void Range::SetOldSites(const Range& src)
{
  //(Called by RecTree)

  // As an optimization, src.m_oldsites may not be set if it
  // has never been relevant; if so our old active sites
  // are simply src's active sites.

  //active/inactive:
  if (!src.m_oldsitesset) {
    m_oldsites = src.m_activesites;
  }
  else {
    m_oldsites = src.m_oldsites;
  }
  m_oldsitesset = true;
  assert(m_oldsites.size() > 0);

  //FC also needs to be updated.
  if (!src.m_oldfcset) {
    m_oldfcsites = src.m_fcsites;
  }
  else {
    m_oldfcsites = src.m_oldfcsites;
  }
  m_oldfcset = true;

  UpdateNewLinks();

} /* Range::SetOldsites */

//_________________________________________________
bool Range::IsSiteActive(long site) const
{
  //This function ignores both FC and alwaysactive, since it's used in
  // likelihood calculations and in branch.cpp for finding valid
  // (i.e. coalescence-information-containing) parents and children.
  if (m_activesites.empty()) return false;

  rangeset::const_iterator range;
  for(range = m_activesites.begin(); range != m_activesites.end(); ++range) {
    if (range->second <= site) continue;

    if (range->first <= site) return true;
    break;
  }

  return false;

} /* Range::IsSiteActive */

//_________________________________________________
void Range::SetCRange(const Range& child1, const Range& child2)
{
  m_activesites = Union(child1.m_activesites,child2.m_activesites);
  SetActiveFCWith(Union(child1.m_fcsites, child2.m_fcsites));

} /* Range::SetCRange */

//_________________________________________________

void Range::SetMRange(const Range& child1)
{
  m_activesites = child1.m_activesites;
  SetActiveFCWith(child1.m_fcsites);

} /* Range::SetMRange */

//_________________________________________________

void Range::SetRRange(const Range& child1, const rangepair& newsites,
                      bool isleft)
{
  rangeset newset;
  newset.insert(newsites);
  m_activesites = Intersection(child1.m_activesites,newset);
  if (isleft) {
    m_side = left_side;
    m_recsite = newsites.second - 1;
  }
  else {
    m_side = right_side;
    m_recsite = newsites.first - 1;
  }
  SetActiveFCWith(child1.m_fcsites);
} /* Range::SetRRange */

//_________________________________________________

void Range::SetFCRange(const rangeset& fc)
{
  SetActiveFCWith(fc);
}

//_________________________________________________

void Range::ClearActiveSites()
{
  m_activesites.clear();
}

//_________________________________________________

bool Range::UpdateCRange(const Range& child1, const Range& child2, bool dofc)
{
  rangeset newsites = Union(child1.m_activesites,child2.m_activesites);
  rangeset newfcsites = Union(child1.m_fcsites, child2.m_fcsites);
  // LS DEBUG  Is this needed?
  newfcsites = Intersection(newfcsites, newsites);

  //Here, doing the check actually might help.
  if (newsites == m_activesites && 
      (!dofc || newfcsites == m_fcsites)) return false;

  SetOldSitesIfNeeded();

  m_activesites = newsites;
  if (dofc) {
    SetActiveFCWith(newfcsites);
  }
  UpdateNewLinks();

  return true;

} /* Range::UpdateCRange */

//_________________________________________________

bool Range::UpdateMRange(const Range& child1, bool dofc)
{
  //LS NOTE:  The following test would theoretically work, but it's a
  // migration!  What's going to change?

  //if ( m_activesites == child1.m_activesites &&
  //    (!dofc || m_fcsites == child1.m_fcsites)) return false;

  SetOldSitesIfNeeded();

  m_activesites = child1.m_activesites;
  if (dofc) {
    SetActiveFCWith(child1.m_fcsites);
  }
  
  UpdateNewLinks();

  return true;

} /* Range::UpdateMRange */

//_________________________________________________

bool Range::UpdateRRange(const Range& child1, bool dofc)
{
  SetOldSitesIfNeeded();

  rangeset recrange;
  if (m_side == left_side)
    recrange.insert(rangepair(0L,m_recsite+1));
  else if (m_side == right_side)
    recrange.insert(rangepair(m_recsite+1,m_totalsites));
  else {
    assert(false); //A recombination without a set side?
    throw implementation_error("A recombination was specified without knowledge of which side of the recombination we were on.");
  }

  rangeset newactive = Intersection(child1.m_activesites,recrange);
  //LS NOTE:  Even in highly-recombinant cases, running this test proved to
  // be a slowdown (though accurate).  So don't bother and just return true.
  //if (newactive == m_activesites && 
  //    (!dofc || m_fcsites == child1.m_fcsites)) return false;
  m_activesites = newactive;
  if (dofc) {
    SetActiveFCWith(child1.m_fcsites);
  }
  UpdateNewLinks();

  return true;

} /* Range::UpdateRRange */

//_________________________________________________

bool Range::UpdateFCToInclude(const rangeset& fc)
{
  //The new 'fc' rangeset contains a list of only the sites that have most
  // recently achieved final coalescence.
  rangeset newsites = Intersection(m_activesites, fc);
  if (newsites.size() == 0) return false;
  newsites = Union(m_fcsites, fc);
  if (m_fcsites == newsites) return false;
  SetOldSitesIfNeeded();
  m_fcsites = newsites;
  UpdateNewLinks();
  return true;
}

//_________________________________________________

void Range::UpdateFCToExclude(const rangeset& nofc)
{
  //This is used when we break a branch, and all active sites on that branch
  // must now be excluded from fc for all rootward branches.
  SetOldSitesIfNeeded();
  rangeset newsites = RemoveRangeFromRange(nofc, m_fcsites);
  SetActiveFCWith(newsites);
  UpdateNewLinks();
}

//_________________________________________________

bool Range::operator==(const Range& other) const
{
  return (m_activesites == other.m_activesites &&
          m_recsite == other.m_recsite);
}

//_________________________________________________

rangeset Range::GetNonFCActiveSites() const
{
  //LS NOTE:  This function is used in three places, all of which do need to
  // include any alwaysactive sites.  The first place is in Prune()--if a branch
  // has nothing but 'alwaysactive' sites in it, it must not be pruned, as its
  // spouse may have been from a different disease partition.
  // We do use FC here because the particular place we're looking at is one
  // half of a recombination event--the FC sites can travel to the root just
  // as easily from the other half of the recombination as this half; we don't
  // need the event here.
  //
  // The second place is in UpdateTimeListToExcludeFromFC(), which is used
  // when breaking a branch to tell the rest of the tree what needs to be newly
  // considered for rec events.  Any site that has reached FC already before
  // the newly-broken branch will still be FC rootward.  But alwaysactive
  // sites from broken branches may indeed induce new recombination events.
  //
  // Finally, it is used in some debugging functions that track FC two different
  // ways and compare the two.  It needs to do the same thing that
  // branch-breaking did, above, so again needs to consider alwaysactive sites.

  return Union(RemoveRangeFromRange(m_fcsites, m_activesites), m_alwaysactive);
}

//_________________________________________________
//_________________________________________________

//The following routines use both activesites and alwaysactivesites.

long Range::GetFirstNonFCActiveSite() const
{
  //This needs to be FC-aware, since it's used to determine where to place
  // a recombination event, and FC sites are not eligible.
  rangeset active = RemoveRangeFromRange(m_fcsites, Union(m_activesites, m_alwaysactive));
  assert (!active.empty());
  return active.begin()->first;
}

//_________________________________________________
bool Range::NoActiveSites() const
{
  //We ignore FC for this.  We allow branches to propagate to the root even if
  // they only contain FC sites, because this way the FC sites make it to the
  // root.  We just don't allow the FC sites to induce recombination events.
  if (!(m_activesites.empty())) {
    return false;
  }
  for (set<rangepair>::const_iterator ri = m_alwaysactive.begin();
       ri != m_alwaysactive.end(); ri++) {
    switch(m_side) {
    case right_side:
      if (m_recsite < (*ri).first) {
        return false;
      }
      break;
    case left_side:
      if ((*ri).first <= m_recsite) {
        return false;
      }
      break;
    case unset_side:
      //For a non-recombination branch, any always-active sites means
      // we always return false from this routine.
      return false;
    }
  }
  //cout << endl << "No active sites:  ";
  //PrintInfo();
  return true;
};

//_________________________________________________

bool Range::IsAlwaysActiveInRange() const
{
  for (set<rangepair>::const_iterator ri = m_alwaysactive.begin();
       ri != m_alwaysactive.end(); ri++) {
    switch(m_side) {
    case right_side:
      if (m_recsite < (*ri).first) {
        //Warning!  Assumes alwaysactives are exactly one site 'wide'. LS DEBUG?
        return true;
      }
      break;
    case left_side:
      if ((*ri).first <= m_recsite) {
        return true;
      }
      break;
    case unset_side:
      //For a non-recombination branch, any always-active sites means
      // we always return true from this routine.
      return true;
    }
  }
  return false;
}
  

long Range::ActiveLinks() const
{
  //We remove fc sites from the active sites, then add back in any always
  // active sites.
  return Span(Union(RemoveRangeFromRange(m_fcsites, m_activesites), m_alwaysactive));
}

//_________________________________________________

bool Range::FCMatches(rangeset truefc) const
{
  //LS NOTE:  Always called within asserts, so we are free to print stuff:
  rangeset truesupressed = Intersection(truefc, m_activesites);
  rangeset currentsupressed = Intersection(m_fcsites, m_activesites);
  if (truesupressed != currentsupressed) {
    cout << "Problem!  These sites are supposed to be supressed: " << ToString(truesupressed) << endl;
    cout << "But these are instead: " << ToString(currentsupressed) << endl;
    cout << "Other information about this range:" << endl;
    PrintInfo();
    assert(false);
  }
    
  return (truesupressed == currentsupressed);
}

//_________________________________________________

void Range::PrintActives() const
{
  cout << "Active sites:  " << ToString(m_activesites) << endl;
} /* Range::PrintActives */

//_________________________________________________

void Range::PrintNewActives() const
{
  if (!m_oldsitesset) return;
  cout << "Formerly active sites:  " << ToString(m_oldsites) << endl;
  cout << "Newly active sites:  " << ToString(RemoveRangeFromRange(m_oldsites, m_activesites)) << endl;

} /* Range::PrintNewActives */

//_________________________________________________

void Range::PrintAlwaysActives() const
{
  cout << "Always active sites:  " << ToString(m_alwaysactive) << endl;

} /* Range::PrintNewActives */

//_________________________________________________

void Range::PrintFC() const
{
  cout << "Sites that have reached final coalescence:  " << ToString(m_fcsites) << endl;

} /* Range::PrintNewActives */

//_________________________________________________

void Range::PrintOldFC() const
{
  if (!m_oldfcset) return;
  cout << "Sites that used to have reached final coalescence:  " << ToString(m_oldfcsites) << endl;

} /* Range::PrintNewActives */

//_________________________________________________

void Range::PrintRightOrLeft() const
{
  cout << "Right or left:  ";
  switch(m_side) {
  case right_side:
    cout << "right";
    break;
  case left_side:
    cout << "left";
    break;
  case unset_side:
    cout << "unset";
    break;
  }
  cout << endl;
}
//_________________________________________________

void Range::PrintInfo() const
{
  PrintActives();
  PrintAlwaysActives();
  PrintNewActives();
  PrintFC();
  PrintOldFC();
  cout << "Number of active links:  " << ActiveLinks() << endl;
  if (m_newlinks > 0) {
    cout << "Number of newly-active links:  " << m_newlinks << endl;
  }
  cout << "Rec site = " << m_recsite << endl;
  PrintRightOrLeft();
}
