// $Id: range.h,v 1.17 2007/09/19 22:24:29 lpsmith Exp $

/* 
   Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

   This software is distributed free of charge for non-commercial use
   and is copyrighted.  Of course, we do not guarantee that the software
   works, and are not responsible for any damage you may cause or have.
*/

// Range implements an ordered set of half-open site ranges (open at the
// end).  This class is meant to be embedded within a branch to allow
// for the possibility of recombination events.
//
// It tracks two different ranges, the active sites and the newly-active
// sites; plus the derived quantities number of active links and number of
// newly-active links.  Finally, for a recombinant branch, it notes
// which site is associated with the recombination event (i.e. the site
// before the cut link).
//
// It provides getter/setter pairs for all of its internal variables.
// 
// Range provides IsSiteActive() which answers whether the passed site
// is active on the owning branch.
//
// Range also provides a set of functions (1 per force) for
// handling active and newly-active site management during rearrangement.
// These functions are currently: 
//    UpdateCRange(), UpdateMRange() and UpdateRRange() plus
//    SetCRange(), SetMRange() and SetRRange().

#ifndef RANGE_H
#define RANGE_H

#include "constants.h"
#include "rangex.h"
#include <algorithm>
#include <assert.h>
#include <iostream>
#include <stdlib.h>

enum side {right_side, left_side, unset_side};

//____________________________________________________________________
//____________________________________________________________________

class Range
{
private:
  Range(); //undefined
  rangeset m_oldsites, m_activesites, m_newactivelinks;
  rangeset m_oldfcsites, m_fcsites;
  long     m_recsite,  m_newlinks;
  rangeset m_alwaysactive;
  side     m_side;
  bool     m_oldsitesset, m_oldfcset;
  long     m_totalsites;
  
  long     Span(const rangeset& sites) const;
  void     SetOldSitesIfNeeded();
  //UpdateNewLinks uses both active and alwaysactive sites.
  void     UpdateNewLinks();
  void     SetActiveFCWith(const rangeset& fc);


public:
  Range(long nsites, const std::vector<long>& alwaysact);
  // we accept default copy ctor, operator= and dtor


  //Functions that don't care about the active sites.
  void ClearOldSites();
  void ClearNewactiveLinks();
  long GetRecsite()                   const {return m_recsite;};
  bool OldSitesAreLeft(long recsite);

  //Functions that involve only the real active sites.
  rangeset GetActiveSites()           const {return m_activesites;};
  void SetOldSites(const Range& src);
  bool IsSiteActive(long site)        const;
  void SetCRange(const Range& father, const Range& mother);
  void SetMRange(const Range& mother);
  void SetRRange(const Range& mother, const rangepair& newsites,
                 bool  ishigh);
  void SetFCRange(const rangeset& fc);
  void ClearActiveSites(); //For RecTree::Prune

  bool UpdateCRange(const Range& father, const Range& mother, bool dofc);
  bool UpdateMRange(const Range& mother, bool dofc);
  bool UpdateRRange(const Range& mother, bool dofc);
  bool UpdateFCToInclude(const rangeset& fc);
  void UpdateFCToExclude(const rangeset& nofc);
  bool SameActive(const rangeset& src)const {return m_activesites == src;};
  bool operator==(const Range& other) const;
  rangeset GetNonFCActiveSites()      const;


  //Functions that involve the active sites plus the 'always active' sites.
  // (This category includes the private function UpdateNewLinks, above)
  rangeset GetNewActiveLinks()        const {return m_newactivelinks;};
  long GetFirstNonFCActiveSite()      const;
  long ActiveLinks()                  const;
  long NewActiveLinks()               const {
    assert(m_newlinks >= 0);
    return m_newlinks;};
  bool NoActiveSites()                const;
  bool IsAlwaysActiveInRange()        const;

  
  //Debugging functions.
  bool FCMatches(rangeset truefc) const;
  void PrintActives() const;
  void PrintNewActives() const;
  void PrintAlwaysActives() const;
  void PrintFC() const;
  void PrintOldFC() const;
  void PrintRightOrLeft() const;
  void PrintInfo() const;
};

#endif /* RANGE_H */

