//$Id: summary.cpp,v 1.36 2007/10/03 18:06:34 erynes Exp $

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "constants.h"
#include "intervaldata.h"
#include "region.h"
#include "summary.h"
#include "force.h"
#include <assert.h>
#include <cmath>
#include <iostream>
using namespace std;

//____________________________________________________________
//____________________________________________________________

const DoubleVec1d& Summary::GetShortPoint() const
{
  return shortpoint;

} /* Summary::ShortPoint */

//____________________________________________________________

const DoubleVec1d& Summary::GetShortWait() const
{
  return shortwait;
} /* Summary::ShortWait */

//____________________________________________________________

const LongVec2d& Summary::GetShortPicks() const
{
  return shortpicks;
} /* Summary::ShortPicks */

//____________________________________________________________

Interval const * Summary::GetLongPoint() const
{
  assert(!shortness);  // tried to get long-form inappropriately?
  return front;
} /* Summary::LongPoint */

//____________________________________________________________

const list<Interval>& Summary::GetLongWait() const
{
  assert(!shortness);  // tried to get long-form inappropriately?
  return intervalData.intervals;
}  /* Summary::LongWait */

//____________________________________________________________

const bool Summary::GetShortness() const {
  return shortness;
}

//____________________________________________________________

void Summary::AddInterval(double time, const LongVec2d& pk,
			  const LongVec1d& xk, long s, xpart_t ostat,
			  xpart_t nstat, long recsite, 
			  const LongVec1d& picks, force_type type)
{
  // add interval, retaining a pointer to it
  Interval* thisinterval = intervalData.
     AddInterval(back, time, pk, xk, s, ostat, nstat, recsite, picks, type);
  // put that pointer in appropriate place(s)
  if (!front) front = thisinterval;
  back = thisinterval;

} /* Summary::AddInterval */

//____________________________________________________________

bool Summary::Compress()
{
  ComputeShortWait();
  ComputeShortPoint();
  ComputeShortPicks();
  return shortness;
} /* Summary::Compress */

//____________________________________________________________
//____________________________________________________________

Summary* CoalSummary::Clone(IntervalData& interval) const
{
  // NB:  This takes a reference to the IntervalData to which
  // the new Summary will belong.  It does *not* copy the
  // contents of the old Summary; only the type and shortness.

  return new CoalSummary(interval, shortness);
} /* Clone */

//____________________________________________________________

void CoalSummary::ComputeShortPoint()
{ 
  // set up the recipient vector
  shortpoint.assign(nparams, 0.0);

  Interval* in;

  for (in = front; in != NULL; in = in->next) 
    {
      ++shortpoint[in->oldstatus];
    }

} /* CoalSummary::ComputeShortPoint */

//____________________________________________________________

void CoalSummary::ComputeShortWait()
{
  // set up the recipient vector
  shortwait.assign(nparams, 0.0);

  list<Interval>::const_iterator interval = intervalData.begin();
  list<Interval>::const_iterator end = intervalData.end();
  double starttime = 0.0;
  
  for ( ; interval != end; ++interval) {
    double deltaTime = (*interval).endtime - starttime;
    long param;
    for (param = 0; param < nparams; ++param) {
      // OPT:  pre-calculate k(k-1) in storage routines....
      double k = (*interval).xpartlines[param];
      shortwait[param] += deltaTime * k * (k - 1);
    }
    starttime = (*interval).endtime;
  }

} /* CoalSummary::ComputeShortWait */

//____________________________________________________________

void CoalSummary::ComputeShortPicks()
{
  const ForceSummary& fs = registry.GetForceSummary();
  long numlocalforces = fs.GetNLocalPartitionForces();

  // if there are no local-partition-forces, do nothing
  if (numlocalforces == 0) return;
  // if there is no recombination, do nothing
  if (!fs.CheckForce(force_REC)) return;
  
  shortpicks.clear();
  const ForceVec& forces = fs.GetPartitionForces();
  unsigned long i;
  long index = 0;
  for (i = 0; i < forces.size(); ++i) {
    if (forces[i]->IsLocalPartitionForce()) {
      const PartitionForce* partforce =
	      dynamic_cast<const PartitionForce*>(forces[i]);
      long nparts = partforce->GetNPartitions();
      LongVec1d counts(nparts, 0L);
      list<Interval>::const_iterator interval = intervalData.begin();
      list<Interval>::const_iterator end = intervalData.end();
      for ( ; interval != end; ++interval) {
        if (!(*interval).partnerpicks.empty()) {
          ++counts[(*interval).partnerpicks[index]];
	}
      }
      shortpicks.push_back(counts);
      ++index;
    }
  }
} /* CoalSummary::ComputeShortPicks */

//____________________________________________________________

void CoalSummary::AdjustSummary(const DoubleVec1d& totals, long region)
{

  Interval* fakefront = NULL;
  Interval* fakeback = NULL;
  assert(totals.size()==shortpoint.size()); // inconsistent input?
  unsigned long nparams(totals.size());
  bool didadjust = false;

  unsigned long param;
  for(param = 0; param < nparams; ++param) {
     if (totals[param] != 0.0) continue;

     if (!shortness) {
       didadjust = true;
       Interval* newinterval = fakeIntervals.AddDummyInterval(fakeback,
          0L, param, FLAGLONG, FLAGLONG, force_COAL);

       if (!fakefront) fakefront = newinterval;
       fakeback = newinterval;
     }

     // adjust short-form summary statistics
     shortpoint[param] += 1.0;
  }

  if (didadjust) {
    // hook fake intervals onto front of real intervals
    fakeback->next = front;
    front = fakefront;
  }

} /* CoalSummary::AdjustSummary */

//____________________________________________________________
//____________________________________________________________

Summary* MigSummary::Clone(IntervalData& interval) const
{
  // NB:  This takes a reference to the IntervalData to which
  // the new Summary will belong.  It does *not* copy the
  // contents of the old Summary; only the type and shortness.

  return new MigSummary(interval, shortness);
} /* Clone */

//____________________________________________________________

void MigSummary::ComputeShortPoint()
{ 
  // set up the recipient vector
  shortpoint.assign(m_npop * m_npop, 0.0);

  Interval* in;

  for (in = front; in != NULL; in = in->next) {
    ++shortpoint[m_npop * in->oldstatus + in->newstatus];
  }

} /* MigSummary::ComputeShortPoint */

//____________________________________________________________

void MigSummary::ComputeShortWait()
{
  // set up the recipient vector
  shortwait.assign(m_npop, 0.0);

  list<Interval>::const_iterator interval = intervalData.begin();
  list<Interval>::const_iterator end = intervalData.end();
  long pop;
  double starttime = 0.0;

  for ( ; interval != end; ++interval) {
    double deltaTime = (*interval).endtime - starttime;
    for (pop = 0; pop < m_npop; ++pop) {
      double k = (*interval).partlines[m_migpartindex][pop];
      shortwait[pop] += deltaTime * k;
    }
    starttime = (*interval).endtime;
  }

} /* MigSummary::ComputeShortWait */

//____________________________________________________________

void MigSummary::AdjustSummary(const DoubleVec1d& totals, long region)
{
  assert(static_cast<xpart_t>(totals.size()) == m_npop*m_npop);
  Interval* fakefront = NULL;
  Interval* fakeback = NULL;
  
  //LS NOTE: It's possible that this will add a migration event for a
  // migration that's been constrained to be zero.  However, the maximizer
  // ignores all such events in this case, so adding one is not fatal.  Though
  // one should be wary.

  bool didadjust = false;
  
  xpart_t frompop, topop;
  
  for (frompop = 0; frompop != m_npop; ++frompop) {
    for (topop = 0; topop != m_npop; ++topop) {

      // no adjustment to diagonal entries
      if (frompop == topop) continue;

      long param = frompop * m_npop + topop;   // index into linearized vector
      if (totals[param] == 0.0) {  // no events of this type
        if (!shortness) {
          didadjust = true;

          // add fake interval
	  Interval* newinterval = 
	    fakeIntervals.AddDummyInterval(fakeback, 0L,
               frompop, topop, FLAGLONG, force_MIG);
          if (!fakefront) fakefront = newinterval;
          fakeback = newinterval;
        }

        // adjust short-form summary statistics
        shortpoint[param] += 1.0;
      }
    }
  }

  if (didadjust) {
    // hook fake intervals onto front of real intervals
    fakeback->next = front;
    front = fakefront;
  }
} /* MigSummary::AdjustSummary */

//____________________________________________________________
//____________________________________________________________

Summary* DiseaseSummary::Clone(IntervalData& interval) const
{
  // NB:  This takes a reference to the IntervalData to which
  // the new Summary will belong.  It does *not* copy the
  // contents of the old Summary; only the type and shortness.

  return new DiseaseSummary(interval, shortness);
} /* Clone */

//____________________________________________________________

void DiseaseSummary::ComputeShortPoint()
{ 
  // set up the recipient vector
  shortpoint.assign(m_nstati * m_nstati, 0.0);

  Interval* in;

  for (in = front; in != NULL; in = in->next) {
    ++shortpoint[m_nstati * in->oldstatus + in->newstatus];
  }

} /* DiseaseSummary::ComputeShortPoint */

//____________________________________________________________

void DiseaseSummary::ComputeShortWait()
{
  // set up the recipient vector
  shortwait.assign(m_nstati, 0.0);

  list<Interval>::const_iterator interval = intervalData.begin();
  list<Interval>::const_iterator end = intervalData.end();
  long status;
  double starttime = 0.0;

  for ( ; interval != end; ++interval) {
    double deltaTime = (*interval).endtime - starttime;
    for (status = 0; status < m_nstati; ++status) {
      double k = (*interval).partlines[m_dispartindex][status];
      shortwait[status] += deltaTime * k;
    }
    starttime = (*interval).endtime;
  }

} /* DiseaseSummary::ComputeShortWait */

//____________________________________________________________

void DiseaseSummary::AdjustSummary(const DoubleVec1d& totals, long region)
{
  assert(static_cast<xpart_t>(totals.size()) == m_nstati*m_nstati);
  Interval* fakefront = NULL;
  Interval* fakeback = NULL;

  bool didadjust = false;
  
  xpart_t oldstatus, newstatus;
  
  for (oldstatus = 0; oldstatus != m_nstati; ++oldstatus) {
    for (newstatus = 0; newstatus != m_nstati; ++newstatus) {

      // no adjustment to diagonal entries
      if (oldstatus == newstatus) continue;

      // index into linearized vector
      long param = oldstatus * m_nstati + newstatus;

      if (totals[param] == 0.0) {  // no events of this type
        if (!shortness) {
          didadjust = true;

          // add fake interval
          Interval* newinterval = 
  	    fakeIntervals.AddDummyInterval(fakeback, 0L,
               oldstatus, newstatus, FLAGLONG, force_DISEASE);
          if (!fakefront) fakefront = newinterval;
          fakeback = newinterval;
        }

        // adjust short-form summary statistics
        shortpoint[param] += 1.0;
      }
    }
  }

  if (didadjust) {
    // hook fake intervals onto front of real intervals
    fakeback->next = front;
    front = fakefront;
  }
} /* DiseaseSummary::AdjustSummary */

//____________________________________________________________
//____________________________________________________________

RecSummary::RecSummary(IntervalData& interval, bool shortform) :
Summary(interval,registry.GetForceSummary().GetNParameters(force_REC),
        shortform),
nrecsbyxpart(registry.GetForceSummary().GetNParameters(force_COAL),0L)
{
long i(0L);
const ForceVec forces(registry.GetForceSummary().GetPartitionForces());
ForceVec::const_iterator force;
for(force = forces.begin(); force != forces.end(); ++force) {
   PartitionForce* pforce(dynamic_cast<PartitionForce*>(*force));
   LongVec1d fparts(pforce->GetNPartitions(),0L);
   nrecsbypart.push_back(fparts);
   assert(pforce->GetPartIndex() == i); // not canonical order?!
   ++i;
}

} /* RecSummary::ctor */

//____________________________________________________________

Summary* RecSummary::Clone(IntervalData& interval) const
{
  // NB:  This takes a reference to the IntervalData to which
  // the new Summary will belong.  It does *not* copy the
  // contents of the old Summary; only the type and shortness.

  return new RecSummary(interval, shortness);
} /* Clone */

//____________________________________________________________

void RecSummary::ComputeShortPoint()
{ 
  // set up the recipient vector
  shortpoint.assign(nparams, 0.0);

  Interval* in;
  for (in = front; in != NULL; in = in->next) {
    ++shortpoint[0];
  }

} /* RecSummary::ComputeShortPoint */

//____________________________________________________________

void RecSummary::ComputeShortWait()
{
  // set up the recipient vector
  shortwait.assign(nparams, 0.0);

  list<Interval>::const_iterator interval = intervalData.begin();
  list<Interval>::const_iterator end = intervalData.end();
  double starttime = 0.0;
  //  long count=1;
  for ( ; interval != end; ++interval) {
    double deltaTime = (*interval).endtime - starttime;
    shortwait[0] += deltaTime * (*interval).activesites;
    starttime = (*interval).endtime;
    //    cout << "ShortWait " << count++ << ": ts=" << (*interval).starttime << " te=" << (*interval).endtime << " actsites=" << (*interval).activesites << " r=" << shortwait[0] << endl; 
  }

} /* RecSummary::ComputeShortWait */

//____________________________________________________________

void RecSummary::AdjustSummary(const DoubleVec1d& totals, long region)
{
  assert(totals.size() == 1);  // inconsistent forces?!
  if (!registry.GetDataPack().GetRegion(region).RecombinationCanBeEstimated()) {
    return;
  }
  Interval* fakefront = NULL;
  Interval* fakeback = NULL;
  bool didadjust = false;
  
  long param = 0;
  if (totals[param] == 0.0) {  // no events of this type
    if (!shortness) {
      didadjust = true;

      // add fake interval
      Interval* newinterval = 
        fakeIntervals.AddDummyInterval(fakeback, 0L, FLAGLONG,
        FLAGLONG, 0L, force_REC);
      if (!fakefront) fakefront = newinterval;
      fakeback = newinterval;
    }

    // adjust short-form summary statistics
    shortpoint[param] += 1.0;
  }
  if (didadjust) {
    // hook fake intervals onto front of real intervals
    fakeback->next = front;
    front = fakefront;
  } 
} /* RecSummary::AdjustSummary */

//____________________________________________________________

void RecSummary::AddToRecombinationCounts(const LongVec1d& membership)
{
  // nrecsbyxpart
  nrecsbyxpart[registry.GetDataPack().GetCrossPartitionIndex(membership)]++;
  // nrecsbypart
  assert(membership.size() == nrecsbypart.size());
  LongVec1d::size_type partforce;
  for (partforce = 0; partforce < membership.size(); ++partforce) {
     nrecsbypart[partforce][membership[partforce]]++;
  }

} /* RecSummary::Compress */

//____________________________________________________________
//____________________________________________________________


void Summary::AddShortPoint(string s)
{
  double a = static_cast<long>(atoi(s.c_str()));
  shortpoint.push_back(a);
}
void Summary::AddShortWait(string s)
{
  double a = static_cast<double>(atof(s.c_str()));
  shortwait.push_back(a);
}
void Summary::AddShortPicks(const StringVec1d& svec)
{
  LongVec1d newpicks;
  StringVec1d::const_iterator values;
  for(values = svec.begin(); values != svec.end(); ++values) {
    long a = static_cast<long>(atoi(values->c_str()));
    newpicks.push_back(a);
  }
  shortpicks.push_back(newpicks);
}
