// $Id: timelist.cpp,v 1.24 2007/09/28 22:00:13 lpsmith Exp $

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "timelist.h"
#include "stringx.h" // for access to Pretty() in Printtimelist()
                     //     also for ToString() in Printtreelist()
#include <set>       // for Printtreelist()
#include "datapack.h" // for TipData class 
#include "errhandling.h" // for exceptions
#include <functional>   // for mem_fun
#include <fstream>    // for WriteBestTree
#include <assert.h>
#include "force.h"    // for SetStickParams() use of StickForce.

using namespace std;

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif
//____________________________________________________________________________

bool is_sorted(Branchconstiter begin, Branchconstiter end)
{
  if (begin==end) return true;
  Branchconstiter i = begin;
  Branchconstiter j = i;
  for (j++; j!=end && i!= end; i++, j++) {
    if ((*j)->m_eventtime < (*i)->m_eventtime) return false;
    if ((*j)->m_eventtime == (*i)->m_eventtime &&
        (*j)->GetID() < (*i)->GetID()) return false;
  }
  return true;
}

//____________________________________________________________________________
// TimeList constructor

TimeList::TimeList() :
m_stickparams(unknown_region)
{
base = Branch_ptr(new BBranch);
base->m_eventtime = DBL_MAX;

m_firsttip = branches.end();
m_firstbody = branches.end();
m_firstcoal = branches.end();

ntips = ncuttable = 0;

} /* TimeList::TimeList */

//____________________________________________________________________________

TimeList::TimeList(const TimeList& src) :
m_stickparams(unknown_region)
{

base = Branch_ptr(new BBranch);
CopyTips(src,true);
CopyBody(src,true);
CopyStick(src);

} /* TimeList copy ctor */

//____________________________________________________________________________

TimeList& TimeList::operator=(const TimeList& src)
{

CopyTips(src,true);
CopyBody(src,true);
CopyStick(src);

return *this;

} /* TimeList::operator= */

//____________________________________________________________________________

TimeList::~TimeList()
{
// empty due to shared_ptr
} /* TimeList::~TimeList */

//____________________________________________________________________________

void TimeList::Clear()
{
branches.clear();

m_firsttip = branches.end();
m_firstbody = branches.end();
m_firstcoal = branches.end();

ntips = ncuttable = 0;
ClearBranchCount();

} /* TimeList::Clear */

//____________________________________________________________________________

void TimeList::ClearLeavingStick()
{
ClearBranchCount();

branches.remove_if(IsBodyGroup());
branches.remove_if(IsTipGroup());

ncuttable = 0;

m_firstbody = branches.end();
m_firstcoal = branches.end();
m_firsttip = branches.end();

} /* TimeList::ClearLeavingStick */

//____________________________________________________________________________

void TimeList::ClearBody()
{
ClearBranchCount();    // clears all branches from counter

Branchiter brit;
for(brit = FirstTip(); brit != End(); brit = NextTip(brit)) {
   (*brit)->SetParent(0,Branch::NONBRANCH);
   (*brit)->SetParent(1,Branch::NONBRANCH);
   CountBranch((*brit)->Event());    // re-introduces tips only
}

branches.remove_if(IsBodyGroup());
ncuttable = GetNTips();

// reset stored locations into body
m_firstbody = branches.end();
m_firstcoal = branches.end();

} /* TimeList::ClearBody */

//____________________________________________________________________________

void TimeList::ClearStick()
{
m_joints.clear();

} /* TimeList::ClearStick */

//____________________________________________________________________________

void TimeList::ClearPartialBody(Branchiter firstinvalid)
{

double cuttime = (*firstinvalid)->m_eventtime;

Branchiter brit;
for (brit = Begin(); brit != firstinvalid; ++brit) {
   Branch_ptr branch = *brit;
   long pa(0L);
   for (pa = 0; pa < NELEM; ++pa) {
// WARNING warning -- does this comparison fail for tied non-tip branches?
      if (branch->Parent(pa)) {
	 // I believe this assert checks for tie-related failure--Mary
	 assert(!(branch->Parent(pa) == *firstinvalid && 
             branch->Parent(pa)->m_eventtime != cuttime));
         if (branch->Parent(pa)->m_eventtime > cuttime)
            branch->SetParent(pa,Branch::NONBRANCH);
      }
   }
}

// uncount the dying branches. 
for (brit = firstinvalid; brit != branches.end(); brit = NextBody(brit)) {
  ncuttable -= (*brit)->Cuttable();
  UncountBranch((*brit)->Event());
}

bool neednewbody(cuttime <= (*m_firstbody)->m_eventtime);
bool neednewcoal(cuttime <= (*m_firstcoal)->m_eventtime);

// erase unwanted entries
branches.erase(std::remove_if(firstinvalid,End(),IsBodyGroup()),End());

if (neednewbody)
   m_firstbody = std::find_if(branches.begin(),branches.end(),IsBodyGroup());
if (neednewcoal)
   m_firstcoal = std::find_if(branches.begin(),branches.end(),IsCoalGroup());

} /* ClearPartialBody */

//____________________________________________________________________________

void TimeList::CopyWithoutDLCells(const TimeList& src)
{
CopyTips(src,false);
CopyBody(src,false);
CopyStick(src);

} /* TimeList::CopyWithoutDLCells */

//____________________________________________________________________________

long TimeList::GetNBodyBranches() const
{

return std::count_if(Begin(),End(),IsBodyGroup());

} /* TimeList::GetNBodyBranches */

//____________________________________________________________________________

TBranch_ptr TimeList::CreateTip(const TipData& tipdata, const vector<LocusCell>& cells, const vector<LocusCell>& movingcells, Range protorange)
{
  TBranch_ptr ptip = TBranch_ptr(new TBranch(tipdata, protorange));
  ptip->SetDLCells(cells);
  ptip->SetMovingDLCells(movingcells);
  Collate(ptip); //This way tips are added the same way everybody is, and
  //               it's ensured they'll be in the right order.
  ++ntips;

// JDEBUG--should this be tip->IsValid()? assert(IsValid());

return ptip;

} /* TimeList::CreateTip */

//____________________________________________________________________________

TBranch_ptr TimeList::GetTip(const string& tipname) const
{
TBranch_ptr ptip;

Branchconstiter brit;
for(brit = FirstTip(); brit != branches.end(); brit = NextTip(brit)) {
   ptip = boost::dynamic_pointer_cast<TBranch>(*brit);
   assert(ptip);
   if (ptip->m_label == tipname) return ptip;
}

throw data_error("Unable to find tip:  " + tipname);

} /* TimeList::GetTip */

//____________________________________________________________________________

void TimeList::CopyTips(const TimeList &src, bool copydlcells)
{
// Clear out a timelist, then copy in tips from another
// timelist.

ClearLeavingStick();
Branchconstiter brit;
for (brit = src.FirstTip(); brit != src.End(); brit = src.NextTip(brit)){
  CollateAndMakeEquiv((*brit)->Clone(copydlcells), *brit);
  CountBranch((*brit)->Event());
}

ntips = src.ntips;
m_firsttip = std::find_if(branches.begin(),branches.end(),IsTipGroup());

} /* TimeList::CopyTips */

//____________________________________________________________________________
// assuming that by "equivalent" we mean, has the same event, parition
// info, eventtime, activesites, recsite and label (if any).

Branchiter TimeList::FindEquivBranch(const Branch_ptr target)
{
  Branchiter myit;
  assert(target);

  for (myit = Begin(); myit != End(); ++myit)
    if (target->IsEquivalentTo(*myit)) return myit;

  assert(false);

  return myit;

} /* TimeList::FindEquivBranch */

//____________________________________________________________________________

// CopyBody() assumes that the tips of src and *this are identical!
// It also assumes that no internal *node* of the tree has 
// both 2 parents and 2 children.
void TimeList::CopyBody(const TimeList &src, bool copydlcells)
{
ClearBody();

Branch_ptr newbranch;
Branchconstiter brit = src.FirstBody();

// if the src tree has no body, we will not attempt to copy it
// nor to hook up the Base.  Given an incomplete tree, we return
// an incomplete tree.  Mary August 2001

if (brit != src.End()) {
   for( ; brit != src.End(); brit = src.NextBody(brit)) {
      newbranch = (*brit)->Clone(copydlcells);
      Branch_ptr srcchild = (*brit)->Child(0);
      Branch_ptr newchild = srcchild->GetEquivBranch().lock();
      assert(newchild);
      newbranch->SetChild(0,newchild);

      if (srcchild->Parent(0) == *brit) {
        newchild->SetParent(0,newbranch);
        srcchild = (*brit)->Child(1);
        if (srcchild) {
          newchild = srcchild->GetEquivBranch().lock();
          assert(newchild);
          newbranch->SetChild(1,newchild);
          newchild->SetParent(0,newbranch);
        }
      } else
        newchild->SetParent(1,newbranch);

      CollateAndMakeEquiv(newbranch, *brit);
   }

   base->SetChild(0,newbranch);
   newbranch->SetParent(0,base);
}

ncuttable = src.ncuttable;
branchmap = src.branchmap;

} /* TimeList::CopyBody */

//____________________________________________________________________________

void TimeList::CopyPartialBody(const TimeList& src, const Branchiter& srcstart,
   Branchiter& mystart)
{
// This differs from CopyBody in that it only copies the part of the 
// TimeList that differs between this and src, as an optimization.

// Possible speedup involving moving the functionality of 
// ClearPartialBody inside the loop over branches below!?
ClearPartialBody(mystart);

Branch_ptr newbranch;
Branchconstiter brit = src.FirstBody();

// if the src tree has no body, we will not attempt to copy it
// nor to hook up the Base.  Given an incomplete tree, we return
// an incomplete tree.  Mary August 2001

if (brit != src.End()) {
  for(brit = srcstart; brit != src.End(); brit = src.NextBody(brit)) {
      bool withdlcells(true);
      newbranch = (*brit)->Clone(withdlcells);
      Branch_ptr srcchild = (*brit)->Child(0);
      Branch_ptr newchild = srcchild->GetEquivBranch().lock();
      assert(newchild);
      newbranch->SetChild(0,newchild);

      if (srcchild->Parent(0) == *brit) {
        newchild->SetParent(0,newbranch);
        srcchild = (*brit)->Child(1);
        if (srcchild) {
          newchild = srcchild->GetEquivBranch().lock();
          assert(newchild);
          newbranch->SetChild(1,newchild);
          newchild->SetParent(0,newbranch);
        }
      } else
        newchild->SetParent(1,newbranch);

      CollateAndMakeEquiv(newbranch, *brit);
  }

  base->SetChild(0,newbranch);
  newbranch->SetParent(0,base);
}
assert(is_sorted(branches.begin(), branches.end()));

ncuttable = src.ncuttable;
branchmap = src.branchmap;

ExtendStickToCoverTree();

} /* TimeList::CopyPartialBody */

//____________________________________________________________________________

void TimeList::CopyStick(const TimeList& src)
{
ClearStick();

// if the src tree has no stick, we will not attempt to copy it
// Given a stickless tree, we return a stickless tree.  Jon February 2006

if (!(src.NoStick())) {
   list<JBranch_ptr>::const_iterator srcjt(src.m_joints.begin());

   bool copydlcells(false); // there are no dlcells in a JBranch

   // setup the first joint.
   Branch_ptr newjoint = (*srcjt)->Clone(copydlcells);
   newjoint->SetChild(0,Branch::NONBRANCH);
   m_joints.push_back(boost::dynamic_pointer_cast<JBranch>(newjoint));
   Branch_ptr prevjoint(newjoint);

   for(++srcjt; srcjt != src.m_joints.end(); ++srcjt) {

      newjoint = (*srcjt)->Clone(copydlcells);
      newjoint->SetChild(0,prevjoint);
      prevjoint->SetParent(0,newjoint);

      m_joints.push_back(boost::dynamic_pointer_cast<JBranch>(newjoint));
      prevjoint = newjoint;
   }
}

m_stickparams = src.m_stickparams;

} /* TimeList::CopyStick */

//____________________________________________________________________________

void TimeList::AddNextJoint(JBranch_ptr here)
{
assert(m_joints.back() == here);
JBranch_ptr nextjoint = here->AppendJoint(m_stickparams);
m_joints.push_back(nextjoint);
} /* TimeList::AddNextJoint */

//____________________________________________________________________________

void TimeList::SetStickParameters(const ForceParameters& fp)
{
m_stickparams = fp;

if (m_joints.empty()) return;

const StickForce& force(registry.GetForceSummary().GetStickForce());
list<JBranch_ptr>::iterator br;
for(br = m_joints.begin(); br != m_joints.end(); ++br) {
   DoubleVec1d newthetas = force.ComputeStickValues((*br)->m_eventtime,fp);
   (*br)->SetThetas(newthetas);
}

} /* TimeList::SetStickParameters */

//____________________________________________________________________________

void TimeList::ExtendStickToCoverTree()
{
if (m_joints.empty() || 
      (m_joints.back()->m_eventtime > branches.back()->m_eventtime))
   return;

double newendtime = branches.back()->m_eventtime;
JBranch_ptr joint(m_joints.back());
while(joint->m_eventtime < newendtime) {
   JBranch_ptr newjoint = joint->AppendJoint(m_stickparams);
   m_joints.push_back(newjoint);
   joint = newjoint;
}

} /* TimeList::ExtendStickToCoverTree */

//____________________________________________________________________________

JBranch_ptr TimeList::NextJointAppendingIfLast(JBranch_ptr start)
{
JBranch_ptr newjoint(start->NextJointAppendingIfLast(m_stickparams));
if (!(newjoint->NextJoint())) {
   m_joints.push_back(newjoint);
}
return newjoint;
} /* TimeList::NextJointAppendingIfLast */

//____________________________________________________________________________

void TimeList::CreateStick(JBranch_ptr first)
{
ClearStick();
m_joints.push_back(first);
} /* TimeList::CreateStick */

//____________________________________________________________________________

void TimeList::ScoreStick(TreeSummary& treesum) const
{
if (m_joints.empty()) return;

DoubleVec2d stickfreqs;
DoubleVec1d sticklengths;
DoubleVec2d stickthetas;
DoubleVec2d stickratios;

DoubleVec1d tipfreqs;

list<JBranch_ptr>::const_iterator br;
// JDEBUG --- don't use this! for(; br != m_joints.end();) {
for(br = m_joints.begin(); (*br)->NextJoint(); ++br) {
   DoubleVec1d thetas((*br)->GetThetas());
   DoubleVec1d freqs(thetas.size());
   DoubleVec1d ratios(thetas.size());
   double zero(0.0);
   stickthetas.push_back(thetas);
   std::transform(thetas.begin(),thetas.end(),freqs.begin(),
      std::bind2nd(std::divides<double>(),
                   std::accumulate(thetas.begin(),thetas.end(),zero)));
   stickfreqs.push_back(freqs);
   double newlength((*br)->BottomTime() - (*br)->m_eventtime);
   sticklengths.push_back(newlength);

   if (br == m_joints.begin()) tipfreqs = freqs;
   ratios = freqs;
   std::transform(ratios.begin(),ratios.end(),tipfreqs.begin(),
      ratios.begin(),std::divides<double>());
   stickratios.push_back(ratios);

}

treesum.SetStickSummary(stickfreqs,sticklengths,stickthetas,stickratios);

} /* TimeList::ScoreStick */

//____________________________________________________________________________
/*  JDEBUG -- not fixed for new timelist paradigm 2006/02/20
  void TimeList::FillTipsWithDataFrom(const vector<vector<LocusCell> >& tcells)
  {
  // JDEBUG--not robust to sequential sampling!
  // --we decrement tip because the tips are put into the branch list
  // in reverse order by TimeList::CreateTip() (it uses push_front())
  vector<vector<LocusCell> >::const_iterator tcell;
  Branchiter tip;
  for(tip = --BeginBody(), tcell = tcells.begin();
  tip != Begin(); --tip, ++tcell) {
  assert(tcell != tcells.end());
  (*tip)->SetDLCells(*tcell);
  }

  assert(tcell != tcells.end());
  (*tip)->SetDLCells(*tcell);   // need to do one more time to catch first
   

  }*/ /* TimeList::FillTipsWithDataFrom */

//____________________________________________________________________________
/*  JDEBUG -- not fixed for new timelist paradigm 2006/02/20
  void TimeList::SetupBodyDLCellsWith(const vector<LocusCell>& loccell)
  {
  Branchiter branch;
  for(branch = BeginBody(); branch != End(); ++branch) {
  (*branch)->SetDLCells(loccell);
  (*branch)->SetUpdateDL();
  }

  } *//* TimeList::SetupBodyDLCellsWith */

//____________________________________________________________________________

bool TimeIsLessThan(Branch_ptr b1, Branch_ptr b2) {
  return (b1->GetTime() < b2->GetTime());
  //This could also check IDs, but it was already too slow.  We would
  // use it for lower_bound, if we found a way to make that faster.
}

Branchiter TimeList::Collate(Branch_ptr newbranch)
{

  Branchiter brit;

  // Degenerate collate into an empty list.  This might not be necessary at all,
  // except that it set 'ncuttable' to 0 and called 'ClearBranchCount'.  Do
  // we need this?  At least it doesn't hurt anything.
  if (branches.empty()) {
    branches.push_front(newbranch);
    brit = branches.begin();  // since we used push_front
    ncuttable = 0;
    ClearBranchCount();
    UpdateBookkeeping(brit);
    return brit;
  }

  //LS NOTE:  This is one way to find where to insert the new branch.  It
  // ended up being slightly slower and it's a bit more obscure, so below is
  // the slightly-faster, more-plodding version.  I've left it here as a
  // template if anyone wants to try using lower_bound or something like it
  // later.
  /*
    Branchiter lowerbound = lower_bound(branches.begin(), branches.end(),
    newbranch, &TimeIsLessThan);
    while (lowerbound != branches.end() &&
    (*lowerbound)->m_eventtime == newbranch->m_eventtime &&
    (*lowerbound)->GetID() < newbranch->GetID()) {
    lowerbound++;
    }
    if (lowerbound != branches.begin()) {
    Branchiter onelower = lowerbound;
    onelower--;
    while (onelower != branches.begin() &&
    (*onelower)->m_eventtime == newbranch->m_eventtime &&
    (*onelower)->GetID() > newbranch->GetID()) {
    onelower--;
    lowerbound--;
    }
    }
  */

  //LS NOTE:  This routine can take up 1/4 of the time spent in rearrangement,
  // (for particularly high values of recombination) so if you see a way to
  // speed it up, take it.  Do note that a version that used 'lower_bound'
  // ended up being a bit slower, sadly.
  Branchiter lowerbound = branches.begin();
  //Iterate to the first branch whose time is greater than or equal to the
  // new branch's time.
  while (lowerbound != branches.end() &&
         (*lowerbound)->m_eventtime < newbranch->m_eventtime) {
    lowerbound++;
  }
  //Iterate to the first branch whose ID is greater than the current branch's
  // ID, assuming we don't go beyond the current time.
  while (lowerbound != branches.end() &&
         (*lowerbound)->m_eventtime == newbranch->m_eventtime &&
         (*lowerbound)->GetID() < newbranch->GetID()) {
    lowerbound++;
  }
   
  brit = branches.insert(lowerbound,newbranch);

  UpdateBookkeeping(brit);

  return brit;

} /* TimeList::Collate */

//____________________________________________________________________________

Branchiter TimeList::Collate(Branch_ptr newbranch, Branch_ptr spouse)
{
  //Used only to add recombination events, which are tied for their times.
  // This is done purely as a time-saving measure for highly-recombinant
  // trees.
  // Currently, the return value is discarded, but that could change--it
  // now points to the iter of where 'newbranch' went in (and not, say, the
  // earlier of the two possibilities).
  Branchiter spousepoint = Collate(spouse);
  if (newbranch->GetID() > spouse->GetID()) {
    //We should add 'newbranch' after 'spouse', not before.
    spousepoint++;
  }
  Branchiter brit = branches.insert(spousepoint, newbranch);
  UpdateBookkeeping(brit);
  return brit;
}

//____________________________________________________________________________

Branchiter TimeList::CollateAndMakeEquiv(Branch_ptr newbranch, Branch_ptr oldbranch)
{
  assert (newbranch != Branch::NONBRANCH);
  assert (oldbranch != Branch::NONBRANCH);
  newbranch->SetEquivBranch(oldbranch);
  oldbranch->SetEquivBranch(newbranch);
  return Collate(newbranch);
}

//____________________________________________________________________________

void TimeList::UpdateBookkeeping(Branchiter newbranch)
{
  ncuttable += (*newbranch)->Cuttable();
  CountBranch((*newbranch)->Event());
  UpdateFirstsIfNecessary(newbranch);
}

//____________________________________________________________________________

void TimeList::Remove(Branch_ptr badbranch)
{
switch(badbranch->BranchGroup()) {
   case bgroupTip:
      if (badbranch == *m_firsttip) m_firsttip = NextTip(m_firsttip);
      break;
   case bgroupBody:
      if (badbranch == *m_firstbody) m_firstbody = NextBody(m_firstbody);
      if (badbranch == *m_firstcoal) m_firstcoal = NextCoal(m_firstcoal);
      break;
   default:
      throw implementation_error("TimeList::Unknown branch group.");
}

branches.remove(badbranch);
ncuttable -= badbranch->Cuttable();
UncountBranch(badbranch->Event());

} /* TimeList::Remove */

//____________________________________________________________________________

Branchiter TimeList::NextTip(Branchiter& it)
{

Branchiter mit(it);
++mit;
return ((mit == End()) ? End() : std::find_if(mit,End(),IsTipGroup()));

} /* TimeList::NextTip */

//____________________________________________________________________________

Branchconstiter TimeList::NextTip(Branchconstiter& it) const
{

Branchconstiter mit(it);
++mit;
return ((mit == End()) ? End() : std::find_if(mit,End(),IsTipGroup()));

} /* TimeList::NextTip */

//____________________________________________________________________________

Branchiter TimeList::NextBody(Branchiter& it)
{

Branchiter mit(it);
++mit;
return ((mit == End()) ? End() : std::find_if(mit,End(),IsBodyGroup()));

} /* TimeList::NextBody */

//____________________________________________________________________________

Branchconstiter TimeList::NextBody(Branchconstiter& it) const
{

Branchconstiter mit(it);
++mit;
return ((mit == End()) ? End() : std::find_if(mit,End(),IsBodyGroup()));

} /* TimeList::NextBody */

//____________________________________________________________________________

Branchiter TimeList::NextCoal(Branchiter& it)
{

Branchiter mit(it);
++mit;
return ((mit == End()) ? End() : std::find_if(mit,End(),IsCoalGroup()));

} /* TimeList::NextCoal */

//____________________________________________________________________________

Branchconstiter TimeList::NextCoal(Branchconstiter& it) const
{

Branchconstiter mit(it);
++mit;
return ((mit == End()) ? End() : std::find_if(mit,End(),IsCoalGroup()));

} /* TimeList::NextCoal */

//____________________________________________________________________________

Branchiter TimeList::PrevBodyOrTip(Branchiter& it)
{
Branchiter mit(it);
--mit;
IsBodyGroup bodycheck;
IsTipGroup tipcheck;
while(!bodycheck(*mit) && !tipcheck(*mit)) --mit;
return mit;

} /* TimeList::PrevBody */

//____________________________________________________________________________

Branchiter TimeList::FindIter(Branch_ptr branch)
{


return ((branches.empty()) ? End() : std::find(Begin(),End(),branch));

} /* TimeList::FindIter */

//____________________________________________________________________________

Branchconstiter TimeList::FindIter(Branch_ptr branch) const
{
Branchconstiter brit;
for(brit = Begin(); brit != End(); ++brit)
   if (*brit == branch) break;  // yes, we compare pointers!

return brit;

} /* TimeList::FindIter */

//____________________________________________________________________________

void TimeList::Prune()
{
// Excise marked branches.
Branchiter brit;
for(brit = FirstBody(); brit != End();) {
   Branch_ptr pbranch = *brit;
   if (pbranch->m_marked) {
      Branch_ptr pparent = pbranch->Parent(0);
      pparent->ReplaceChild(pbranch,pbranch->Child(0));
      SetUpdateDLs(pparent);

      pparent = pbranch->Parent(1);
      if (pparent) {
         pparent->ReplaceChild(pbranch,pbranch->Child(0));
         SetUpdateDLs(pparent);
      }
      brit = NextBody(brit);
      Remove(pbranch);  // deletes what pbranch points to
      continue;
   }
   brit = NextBody(brit);
}
assert(IsValid());

// Check for loops in the root.
long nbranches = 2*ntips;
for(brit = FirstBody(); ;brit = NextBody(brit)) {
   nbranches += (*brit)->CountDown();
   if (nbranches <= 2) break;
}
if (*brit != Root()) {
   base->SetChild(0,*brit);
   (*brit)->SetParent(0,base);
   (*brit)->SetParent(1,Branch::NONBRANCH);

   // we could use std::remove_if, but we'd still need to do
   // this handloop for stuff like Uncount, and updating of 
   // m_firstfoo pointers.
   for(brit = NextBody(brit); brit != branches.end(); ) {
      Branch_ptr pbranch = *brit;
      brit = NextBody(brit);
      Remove(pbranch);
   }
}
assert(IsValid());

} /* TimeList::Prune */

//____________________________________________________________________________

void TimeList::SetUpdateDLs(Branch_ptr pBranch)
{

if (pBranch->Event() == btypeBase) return;  // the base terminates recursion

if (!pBranch->GetUpdateDL()) {
   pBranch->SetUpdateDL();
   SetUpdateDLs(pBranch->Parent(0));

   if (pBranch->Parent(1))
      SetUpdateDLs(pBranch->Parent(1));
}

} /* TimeList::SetUpdateDLs */

//____________________________________________________________________________

void TimeList::SetAllUpdateDLs()
// marks the entire tree as needing updating--a debug function.
{
  Branchiter brit;
  for (brit = branches.begin(); brit != branches.end(); ++brit)
    (*brit)->SetUpdateDL();
  --brit;
  (*brit)->SetUpdateDL();

} /* SetAllUpdateDLs() */

//____________________________________________________________________________

void TimeList::ClearUpdateDLs()
{
Branchiter brit;
for(brit = branches.begin(); brit != branches.end(); ++brit)
   (*brit)->ClearUpdateDL();

} /* TimeList::ClearUpdateDLs */

//____________________________________________________________________________

void TimeList::Printtimelist()
{

cout << endl;
cout << " New tree start " << endl;
vector<Branch_ptr> endbranch;
vector<Branch_ptr>::iterator br;
Branchiter branch;
double tyme = 0.0;
bool newinterval = true;
for(branch = Begin(); branch != End();) {
   if (newinterval) cout << Pretty(tyme) << ": branch(s) ";
   if ((*branch)->m_eventtime == tyme) {
      cout << (*branch)->GetID();
      cout << (*branch)->Event() << endl;
      (*branch)->m_range.PrintActives();
      (*branch)->m_range.PrintNewActives();
      cout << (*branch)->GetUpdateDL();
      cout << " ";
      long site;
      cout << endl;
      if ((*branch)->Event() == btypeTip) {
        for(site = 0; site < 4; ++site) {
          DNACell* pcell = dynamic_cast<DNACell*>((*branch)->GetDLCell(0, 0, false).get());
          cout << "   " << Pretty(pcell->GetSiteDLs(site)[0][baseA]);
          cout << "   " << Pretty(pcell->GetSiteDLs(site)[0][baseC]);
          cout << "   " << Pretty(pcell->GetSiteDLs(site)[0][baseG]);
          cout << "   " << Pretty(pcell->GetSiteDLs(site)[0][baseT]);
          cout << endl;
        }
      }
      newinterval = false;
   } else {
      cout << "start" << endl;
      newinterval = true;
      if (!endbranch.empty()) {
         cout << "      branch(s) ";
         for(br = endbranch.begin(); br != endbranch.end(); ++br) {
            if (*br != Branch::NONBRANCH) {
               cout << (*br)->GetID();
               cout << (*br)->Event() << endl;
               (*br)->m_range.PrintActives();
               (*br)->m_range.PrintNewActives();
               cout << (*br)->GetUpdateDL();
               cout << " ";
            }
         }
         cout << " end" << endl;
      }
      endbranch = (*branch)->GetChildren();
      tyme = (*branch)->m_eventtime;
      continue;
   }
   ++branch;
}

// catch the branches that end at the root
cout << endl << "      branch(s) ";
for(br = endbranch.begin(); br != endbranch.end(); ++br) {
   if (*br != Branch::NONBRANCH) {
      cout << (*br)->GetID();
      cout << (*br)->Event();
      (*br)->m_range.PrintActives();
      (*br)->m_range.PrintNewActives();
      cout << (*br)->GetUpdateDL();
      cout << " ";
   }
}
cout << " end at the root" << endl;

} /* TimeList::Printtimelist */

//____________________________________________________________________________

void TimeList::Printtreelist() const
{
set<string> branches;

Branchconstiter branch = FirstTip();
cout << ToString((*branch)->GetID());
cout << ToString((*branch)->Event()) << "/" << (*branch)->m_eventtime << ":";
for( ; branch != End(); branch = NextTip(branch)) {
   string id = ToString((*branch)->GetID());
   branches.insert(id);
   cout << " " << id;
}
cout << endl;

for(branch = FirstBody(); branch != End(); branch = NextBody(branch)) {
   string id = ToString((*branch)->GetID());
   branch_type event = (*branch)->Event();
   //cout << id << ToString(event) << "/" << (*branch)->m_eventtime;
   cout << id << ":" << ToString((*branch)->GetUpdateDL()) << "/" << (*branch)->m_eventtime;

   string chid = ToString((*branch)->Child(0)->GetID());
   branches.erase(chid);
   branches.insert(id);

   switch(event) {
   case btypeCoal:
     cout << ":";
     chid = ToString((*branch)->Child(1)->GetID());
     branches.erase(chid);
     break;
   case btypeRec:
     ++branch;
     id = ToString((*branch)->GetID());
     event = (*branch)->Event();
     cout << "/" << id << event << ":";
     branches.insert(id);
     break;
   case btypeMig:
     cout << ":";
     break;
   case btypeDisease:
     cout << ":";
     break;
   case btypeBase:
   case btypeTip:
   case btypeStick:
     assert(false); // unknown branch type
   }
   set<string>::iterator br;
   for(br = branches.begin(); br != branches.end(); ++br)
      cout << " " << *br;
   if (!(*branch)->m_partitions.empty())
      cout << " " << "status = " << (*branch)->m_partitions[0];
   cout << endl;
}

cout << "End of Tree" << endl;

} /* TimeList::Printtreelist() */

//____________________________________________________________________________

void TimeList::PrintIDs() const
{
  for (Branchconstiter brit = branches.begin(); brit != branches.end(); brit++){
    cout << (*brit)->GetID() << "\t"
         << (*brit)->GetTime() << "\t"
         << ToString((*brit)->Event()) << endl;
  }
}
  
//____________________________________________________________________________

void TimeList::Printtips()
{
cout << endl;
cout << " The tips are " << endl;
Branchiter branch;
for(branch = FirstTip(); branch != End(); branch = NextTip(branch)) {
   TBranch_ptr tip = boost::dynamic_pointer_cast<TBranch>(*branch);
   assert(tip);
   cout << "   " << tip->m_label << " with id#";
   cout << tip->GetID();
   if (tip->m_partitions.size()) {
     cout << " " << tip->m_partitions[0];
   }
   cout << endl;
}
cout << "End tips" << endl;


} /* TimeList::Printtips */

//____________________________________________________________________________

void TimeList::Printstick()
{
cout << endl;
list<JBranch_ptr>::iterator joint;
for(joint = m_joints.begin(); joint != m_joints.end(); ++joint) {
   cout << (*joint)->m_eventtime << ":  thetas ";
   DoubleVec1d ths((*joint)->GetThetas());
   DoubleVec1d::iterator th;
   for(th = ths.begin(); th != ths.end(); ++th)
      cout << " " << *th;
   cout << endl;
}

} /* TimeList::Printstick */

//____________________________________________________________________________

void TimeList::OutputJointsToFile() 
// a debug function
{
ofstream of;
of.open("joints",ios::app);
cout << "New tree" << endl;

list<JBranch_ptr>::iterator joint;
for(joint = m_joints.begin(); joint != m_joints.end(); ++joint) {
	cout << (*joint)->m_eventtime << "  ";
	DoubleVec1d thetas = (*joint)->GetThetas();
        unsigned long i;
	for (i = 0; i < thetas.size(); ++i) {
            cout << thetas[i] << "  ";
	}
    cout << endl;
}
of.close();

} /* TimeList::OutputJointsToFile */

//____________________________________________________________________________

long TimeList::CountJoints() const
{
return m_joints.size();

} /* TimeList::Countjoints */

//____________________________________________________________________________

void TimeList::MakeCoalescent(double theta)
{
assert (HowMany(btypeMig) == 0 && HowMany(btypeRec) == 0);

long nbranches = GetNTips();

Branchiter branch;
for(branch = FirstBody(); branch != End(); branch = NextBody(branch)) {
   assert(nbranches > 0);

   (*branch)->m_eventtime = theta/(nbranches * (nbranches-1));
   --nbranches;
}

} /* TimeList::MakeCoalescent */

//____________________________________________________________________________

string TimeList::DLCheck(const TimeList& other) const
{

// This code should be commented in if DLCheck is being used to test
// for strict identity, and commented out if you know the trees are
// not in fact identical but want to find out where.
#if 1
long ncoals = HowMany(btypeCoal), otherncoals = other.HowMany(btypeCoal);
if (ncoals != otherncoals)
   return string("The trees differ in number of coalescences!\n");
#endif

string problems;
Branchconstiter branch, otherbranch;
for(branch = Begin(), otherbranch = other.Begin();
    branch != End() && otherbranch != other.End();
    ++branch, ++otherbranch) {
   problems += (*branch)->DLCheck(**otherbranch);
}

return problems;

} /* TimeList::DLCheck */

//____________________________________________________________________________

void TimeList::CloneCheck() const
{

vector<Branch_ptr> newbranches;
Branchconstiter branch;
for(branch = Begin(); branch != End(); ++branch) {
   Branch_ptr newbranch = Branch::NONBRANCH;
   bool withdlcells(true);
   newbranch = (*branch)->Clone(withdlcells);
   if (!newbranch) {
      cout << "Failed to clone branch " << (*branch)->GetID() << endl;
   }
   newbranches.push_back(newbranch);
}

} /* TimeList::CloneCheck */

//____________________________________________________________________________

bool TimeList::IsPresent(const Branchiter& branch) const
{

bool foundequiv = false;
Branchconstiter brit;
for(brit = Begin(); brit != End(); ++brit) {
   if (*brit == *branch)
      return true;
   foundequiv = (**brit == **branch);
}

if (foundequiv) 
   cout << "Found equivalent but not same pointer!" << endl;

return false;

} /* TimeList::IsPresent() */

//____________________________________________________________________________

bool TimeList::IsSameExceptForTimes(const TimeList& other) const
{
Branchconstiter brit, otherbrit;
for(brit = Begin(), otherbrit = other.Begin(); brit != End();
   ++brit, ++otherbrit) {
   if (!(*brit)->IsSameExceptForTimes(*otherbrit)) return false;
}

return true;

} /* TimeList::IsSameExceptForTimes */

//____________________________________________________________________________

void TimeList::WriteIntervalLengthsToFile() const
{
ofstream fs;
fs.open("timeints.out",ios::app);

long interval;
Branchconstiter top, bottom;
for(top = Begin(), bottom = FirstBody(), interval = 0; 
    bottom != End(); 
    top = bottom, bottom = NextBody(bottom),  ++interval) {
    if ((*top)->Event() == btypeRec) {
       top = bottom;
       bottom = NextBody(bottom);
    }
    double length((*bottom)->m_eventtime - (*top)->m_eventtime);
    fs << "int" << interval << " " << length << endl;
}

fs.close();
} /* TimeList::WriteIntervalLengthsToFile */

//____________________________________________________________________________

//____________________________________________________________________________
// The following functions manage the counters for branches of each
// type.
//____________________________________________________________________________

void TimeList::CountBranch(const branch_type tag) 
{
  BranchMap::iterator it = branchmap.find(tag);
  if (it == branchmap.end()) {     // entry not found
    branchmap.insert(make_pair(tag, 1L));
  } else {
    (*it).second++;
  }

} /* CountBranch */

//____________________________________________________________________________

void TimeList::UncountBranch(const branch_type tag)
{
  BranchMap::iterator it = branchmap.find(tag);
  assert(it != branchmap.end());  // why are we trying to remove
                                  // a branch type that isn't there?
  (*it).second--;
  assert((*it).second >= 0);      // why are we trying to remove
                                  // a branch that isn't there?
} /* UncountBranch */

//____________________________________________________________________________

void TimeList::ClearBranchCount()
{
// we don't get rid of entries, since they may be useful
// later when the same kind of branch comes round again.

  BranchMap::iterator it = branchmap.begin();
  BranchMap::iterator end = branchmap.end();

  for ( ; it != end; ++it) {
    (*it).second = 0L;
  }

} /* ClearBranchCount */

//____________________________________________________________________________

JBranch_ptr TimeList::FindStickAtTime(double time) const
{
	// MDEBUG could be much quicker using Parent(0) rather than iterator
  list<JBranch_ptr>::const_iterator joint;
  for(joint = m_joints.begin(); joint != m_joints.end(); ++joint) {
    if ((*joint)->m_eventtime == time) return *joint;
    if ((*joint)->m_eventtime < time && (*joint)->BottomTime() > time)
       return *joint;
  }

  string error("Couldn't find a jointed-stick at time ");
  error += ToString(time) + " or later";
  throw implementation_error(error);

  return boost::dynamic_pointer_cast<JBranch>(Branch::NONBRANCH);
} /* FindStickAtTime */

//_________________________________________________________________________

void TimeList::UpdateFirstsIfNecessary(Branchiter& branch)
{
    switch((*branch)->BranchGroup()) {
       case bgroupTip:
          m_firsttip = std::find_if(branches.begin(),branches.end(),
             IsTipGroup());
          return;
       case bgroupBody:
	   if (m_firstbody == branches.end() || 
              (*m_firstbody)->m_eventtime > (*branch)->m_eventtime) {
              m_firstbody = branch;
	   } else {
              if ((*m_firstbody)->m_eventtime == (*branch)->m_eventtime) {
                 m_firstbody = std::find_if(branches.begin(),
                    branches.end(),IsBodyGroup());
	      }
	   }
	   if ((*branch)->Event() == btypeCoal) {
              if (m_firstcoal == branches.end() ||
                 (*m_firstcoal)->m_eventtime > (*branch)->m_eventtime)
                 m_firstcoal = branch;
	      else {
	         if ((*m_firstcoal)->m_eventtime == (*branch)->m_eventtime)
                      m_firstcoal = std::find_if(branches.begin(),
                         branches.end(),IsCoalGroup());
	      }
	   }
           return;
       case bgroupStick:
          assert(false); // shouldn't be any sticks in branchlist anymore
    }
    throw implementation_error("TimeList::Unknown branch group.");
} /* UpdateFirstsIfNecessary */

//____________________________________________________________________________

long TimeList::HowMany(branch_type tag) const
{
  BranchMap::const_iterator it = branchmap.find(tag);
  if (it == branchmap.end()) // entry not found
    return 0L;
  else return (*it).second;

} /* HowMany */

//____________________________________________________________________________

bool TimeList::ContainsOnlyTipsAndCoals() const
{
// trees must always contain tips and coalescences...
if (branchmap.size() > 2) return false;

return true;

} /* ContainsOnlyTipsAndCoals */

//____________________________________________________________________________

bool TimeList::operator==(const TimeList& src) const
{
// we do not check for equivalence of m_firstbody, m_firstcoal,
// and m_firsttip

  if (branchmap != src.branchmap) return false;
  if (ntips != src.ntips) return false;
  if (ncuttable != src.ncuttable) return false;

  if (branches.size() != src.branches.size()) return false;

  Branchconstiter mybr = branches.begin();
  Branchconstiter srcbr = src.branches.begin();

  for ( ; mybr != branches.end(); ++mybr, ++srcbr) {
    if (**mybr != **srcbr) return false;
  }

  return true;

} /* operator== */

//____________________________________________________________________________

Branchiter TimeList::ResolveTiedTimes(Branchconstiter firstinvalid)
{
// the general case
double firsttime((*firstinvalid)->m_eventtime);
Branchiter br;
for(br = FirstBody(); br != End(); br = NextBody(br)) {
   if ((*br)->Event() == btypeStick) continue;
   if ((*br)->m_eventtime == firsttime) return br;
}

assert(false);  // ResolveTiedTimes asked for a time that's not here!

return End();

} /* ResolveTiedTimes */

//____________________________________________________________________________

bool TimeList::IsValid() const
{
  Branchconstiter it;
  long i = 0;
  bool okay = true;

  for (it = Begin(); it != End(); ++it, ++i) {
    if (*it == Branch::NONBRANCH) okay = false; // are there any null branches in this tree?
    else {
      if ((*it)->m_range.NoActiveSites()) {
// JDEBUG
        cout << "A " << ToString((*it)->Event()) << " branch with ";
        (*it)->m_range.PrintActives();
        if ((*it)->Event() == btypeRec) 
           cout << "  and recsite at " << (*it)->m_range.GetRecsite();
        cout << endl;
        const Branch_ptr br((*it)->Child(0));
        cout << "  ActiveSites in child:";
        br->m_range.PrintActives();
        cout << "  ActiveLinks in child:" << br->m_range.ActiveLinks() << endl;
        okay = false;
        assert(false);
      }
      branch_type event = ((*it)->Event());
      switch (event) {
      case btypeTip:
        if (i >= ntips) okay = false;        
        break;
      default:
        if (i < ntips) okay = false;
      }
      assert(okay);

      Branch_ptr parent1 = (*it)->Parent(0);
      Branch_ptr parent2 = (*it)->Parent(1);
      if (parent2 && !parent1) okay = false;
      if (parent1 && !parent1->ConnectedTo(*it)) okay = false;
      if (parent2 && !parent2->ConnectedTo(*it)) okay = false;

      Branch_ptr child1 = (*it)->Child(0);
      Branch_ptr child2 = (*it)->Child(1);
      if (child2 && !child1) okay = false;
      if (child1 && !child1->ConnectedTo(*it)) okay = false;
      if (child2 && !child2->ConnectedTo(*it)) okay = false;
      switch (event) {
      case btypeCoal:
        if ((!(*it)->HasSamePartitionsAs(child1)) || (!(*it)->HasSamePartitionsAs(child2))) {
          okay = false;
          assert(false);
        }
        break;
      case btypeMig:
      case btypeDisease:
        if ((*it)->HasSamePartitionsAs(child1)) {
          okay = false;
          assert(false);
        }
        break;
      case btypeRec:
        if ((*it)->m_range.GetRecsite() == FLAGLONG) okay = false;
        if (!((*it)->PartitionsConsistentWith(child1))) okay = false;
        assert(okay);
        break;
      case btypeStick:
	assert(child1 || (*it)->Parent(0));
	break;
      case btypeBase:
      case btypeTip:
        //Check anything here?
        break;
      }
    }
  }

  assert(m_firsttip == std::find_if(branches.begin(),branches.end(),IsTipGroup()));
  assert(m_firstbody == std::find_if(branches.begin(),branches.end(),IsBodyGroup()));
  assert(m_firstcoal == std::find_if(branches.begin(),branches.end(),IsCoalGroup()));
  assert(okay);
  return okay;

} /* IsValid */

//____________________________________________________________________________
