// $Id: timelist.h,v 1.11 2007/09/28 22:00:13 lpsmith Exp $

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

// This files defines a time ordered container of Branch_ptr,
// TimeList.  The TimeList is a wrapped std::list of all Branch
// derived things associated with a particular tree. 
// Because of the use of shared_ptr to hold Branch objects,
// no deletes should ever be needed.

#ifndef TIMELIST_H
#define TIMELIST_H

#include <vector>
#include "vectorx.h"
#include <map>
#include "defaults.h"
#include "definitions.h"
#include "types.h"
#include "branch.h"

// #include "branch.h"--to create the "base" branch, TimeList constructor
//                      to create a Tbranch, CreateTip
//                      to initialize branch innards, both of the above
//                      to maintain tree child & parent relationships
//                         in CopyBody()
//                      to maintain tree child & parent relationships
//                         in CopyPartialBody()
//                      to track ncuttable via branch.Cuttable()
//                      to track marked status via branch.marked,
//                         branch.SetUpdateDL()

class TipData;
class DataPack;
typedef std::map<branch_type, long> BranchMap;
typedef Branchlist::const_reverse_iterator       Branchconstriter;

//_______________________________________________________________________

class TimeList
{
private:
Branch_ptr base;
Branchlist branches;                 // TimeList owns the branches!!!
BranchMap branchmap;                 // counts branches of each type

// cached start-points, for speed
Branchiter m_firsttip;
Branchiter m_firstbody;
Branchiter m_firstcoal;

// cached force parameter values for stick extension
ForceParameters m_stickparams;

long ntips, ncuttable;

void CountBranch(const branch_type tag);
void UncountBranch(const branch_type tag);
void ClearBranchCount();

// the argument is non-const since it may be assigned to a non-const
// member
void UpdateFirstsIfNecessary(Branchiter& brit);

std::list<JBranch_ptr> m_joints;  // this is owning

public:

TimeList();
TimeList(const TimeList& src);
TimeList& operator=(const TimeList& src);
~TimeList();
void Clear();
void ClearLeavingStick();
void ClearBody();
void ClearPartialBody(Branchiter firstinvalid);
void CopyWithoutDLCells(const TimeList& src); // destroys previous elements
                                              // "copies" tree linkage as well
void CopyTips(const TimeList& src, bool copydlcells);  // destroys previous
                                                       // elements
void CopyBody(const TimeList& src, bool copydlcells);  // destroys previous
                                                       // elements,
                                     // "copies" tree linkage as well,
                                     // assumes tips are already identical
void CopyPartialBody(const TimeList& src, const Branchiter& srcstart,
                     Branchiter& mystart);
                                     // destroys previous elements from
				     // start on, "copies" tree linkage as well,
                                     // assumes tips are already identical

Branchiter FindEquivBranch(const Branch_ptr target);

//void FillTipsWithDataFrom(const vector<vector<LocusCell> >& tcells);
//void SetupBodyDLCellsWith(const vector<LocusCell>& loccell);
long HowMany(const branch_type tag) const;  // how many branchs of type 'tag'?
bool ContainsOnlyTipsAndCoals() const; // any branches other than tips & coals?

Branchiter Collate(Branch_ptr newbranch);
Branchiter Collate(Branch_ptr newbranch, Branch_ptr spouse);
Branchiter CollateAndMakeEquiv(Branch_ptr newbranch, Branch_ptr oldbranch);
void UpdateBookkeeping(Branchiter newbranch);
void Remove(Branch_ptr badbranch);

Branchiter       Begin()           {return branches.begin();};
Branchconstiter  Begin()     const {return branches.begin();};
Branchiter       End()             {return branches.end();};
Branchconstiter  End()       const {return branches.end();};
Branchconstriter RBegin()    const {return branches.rbegin();};
Branchconstriter REnd()      const {return branches.rend();};

// return the first branch of a specific sort
Branchiter       FirstTip() {return m_firsttip;};
Branchconstiter  FirstTip() const {return m_firsttip;};
Branchiter       FirstBody() {return m_firstbody;};
Branchconstiter  FirstBody() const {return m_firstbody;};
Branchiter       FirstCoal() {return m_firstcoal;};
Branchconstiter  FirstCoal() const {return m_firstcoal;};

// return the next branch of a specific sort
Branchiter       NextTip(Branchiter& it);
Branchconstiter  NextTip(Branchconstiter& it) const;
Branchiter       NextBody(Branchiter& it);
Branchconstiter  NextBody(Branchconstiter& it) const;
Branchiter       NextCoal(Branchiter& it);
Branchconstiter  NextCoal(Branchconstiter& it) const;
Branchiter       PrevBodyOrTip(Branchiter& it);

Branch_ptr       Base()            {return base;};
const Branch_ptr Base()      const {return base;};
Branch_ptr       Root()      const {return base->Child(0);};
double           RootTime()  const {return Root()->m_eventtime;};
Branchiter       FindIter(Branch_ptr br);
Branchconstiter  FindIter(Branch_ptr br) const;

long          GetNTips()     const {return ntips;};
long          GetNCuttable() const {return ncuttable;};
long          GetNBranches() const {return branches.size();};
long          GetNBodyBranches() const;
TBranch_ptr   CreateTip(const TipData& tipdata, const vector<LocusCell>& cells,
                        const vector<LocusCell>& movingcells, Range protorange);
TBranch_ptr   GetTip(const string& name) const;
void          Prune();
void          SetUpdateDLs(Branch_ptr);
void          ClearUpdateDLs();
void          SetAllUpdateDLs();

bool operator==(const TimeList& src) const;
bool operator!=(const TimeList& src) const { return !(*this == src); };

Branchiter ResolveTiedTimes(Branchconstiter firstinvalid);

// stick specific functionality
bool             NoStick() const {return m_joints.empty();};
void             ClearStick();
void             CopyStick(const TimeList& src);
// FindStickAtTime throws an implementation error if it fails to find
JBranch_ptr      FindStickAtTime(double time) const;
void             AddNextJoint(JBranch_ptr here);
void             SetStickParameters(const ForceParameters& fp);
void             ExtendStickToCoverTree();
JBranch_ptr      NextJointAppendingIfLast(JBranch_ptr start);
void             CreateStick(JBranch_ptr first);
void             ScoreStick(TreeSummary& treesum) const;
long             CountJoints() const;


// debug functions
void Printtimelist();
void Printtreelist() const;
void PrintIDs() const;
void Printtips();
void Printstick();
bool IsValid() const;
void MakeCoalescent(double theta);
string DLCheck(const TimeList& other) const;
void CloneCheck() const;
bool IsPresent(const Branchiter& branch) const;
bool IsSameExceptForTimes(const TimeList& other) const;
void WriteIntervalLengthsToFile() const;
void OutputJointsToFile();

};

#endif /* TIMELIST_H */
