// $Id: tree.cpp,v 1.83 2007/09/28 22:00:13 lpsmith Exp $

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

#include "branchbuffer.h"  // for Tree::Groom
#include "datapack.h"
#include "dlcell.h"
#include "errhandling.h" // for datalikenorm_error in Tree::CalculateDataLikes
#include "locus.h"
#include "mathx.h" // for exp()
#include "registry.h"
#include "runreport.h"
#include "tree.h"
#include "treesum.h"

#include <algorithm>  // for std::max and min
#include <assert.h>
#include <fstream> 
#include <limits> // for numeric_limits<double>.epsilon()


#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

//#define TEST // erynes test

using namespace std;

Tree::Tree()
: m_overallDL(FLAGDOUBLE),
  m_protorange(NULL),
  m_randomSource(&(registry.GetRandom())),
  m_totalsites(0)
{

}

//____________________________________________________________________________

Tree::~Tree()
{
  delete m_protorange;
  // deliberately blank due to shared_ptr auto cleanup

} /* Tree destructor */

//_____________________________________________________________________________
// NB This copy constructor does *not* copy m_firstinvalid
// (branch equivalence problem needs to be solved).

Tree::Tree(const Tree &tree, bool makestump)
: m_protoCells(tree.m_protoCells),
  m_individuals(tree.m_individuals),
  m_overallDL(tree.m_overallDL),
  m_randomSource(tree.m_randomSource),
  m_pLocusVec(tree.m_pLocusVec),
  m_totalsites(tree.m_totalsites)
{
  m_protorange = new Range(m_totalsites, registry.GetForceSummary().GetSelectedSites());
  
  if (!makestump) {
    // also need to deep copy the branchlist
    m_timelist = tree.m_timelist;

    // now setup the tip pointers in the individuals correctly, since the
    // individual copy ctor only makes empty containers of tips.
    unsigned long indiv;
    for(indiv = 0; indiv < m_individuals.size(); ++indiv) {
      StringVec1d tipnames = tree.m_individuals[indiv].GetAllTipNames();
      vector<Branch_ptr> itips = GetTips(tipnames);
      m_individuals[indiv].SetTips(itips);
    }
  }

}  /* Tree copy ctor */

//____________________________________________________________________________

Tree* Tree::CloneWithoutDLCells() const
{
  Tree* clone = MakeStump();

  bool withnodlcells(true);
  clone->GrowStumpAs(*this,withnodlcells);

  return clone;

} /* Tree::CloneWithoutDLCells */

//____________________________________________________________________________

void Tree::Clear()
{
  m_timelist.Clear();
}

//____________________________________________________________________________

const vector<LocusCell>& Tree::CollectCells()
{
  if (m_protoCells.empty()) {
    unsigned long i;
    for (i = 0; i < m_pLocusVec->size(); ++i) {
      m_protoCells.push_back((*m_pLocusVec)[i].GetProtoCell());
    }
  }
  return m_protoCells;

} /* CollectCells */

//____________________________________________________________________________

void Tree::CopyTips(const Tree *tree)
{
  m_timelist.CopyTips(tree->m_timelist, true);
  m_firstinvalid = m_timelist.Begin();
  m_totalsites = tree->m_totalsites;
  delete m_protorange;
  m_protorange = new Range(m_totalsites, registry.GetForceSummary().GetSelectedSites());

  m_overallDL = FLAGDOUBLE;   // it won't be valid anymore

  m_individuals = tree->m_individuals;
  // now setup the tip pointers in the individuals correctly
  unsigned long indiv;
  for(indiv = 0; indiv < m_individuals.size(); ++indiv) {
     StringVec1d tipnames = tree->m_individuals[indiv].GetAllTipNames();
     vector<Branch_ptr> itips = GetTips(tipnames);
     m_individuals[indiv].SetTips(itips);
  }
  
  vector<Locus>::const_iterator locus;
  m_aliases.clear();
  for(locus = m_pLocusVec->begin(); locus != m_pLocusVec->end(); ++locus)
     m_aliases.push_back(locus->GetDLCalc()->RecalculateAliases(*this, *locus));

} 

//______________________________________________________________________________

void Tree::CopyBody(const Tree *tree)
{
  m_timelist.CopyBody(tree->m_timelist,true);
  m_overallDL = tree->m_overallDL;
} 

//______________________________________________________________________________

void Tree::CopyStick(const Tree* tree)
{
  m_timelist.CopyStick(tree->m_timelist);
}

//______________________________________________________________________________

void Tree::CopyPartialBody(const Tree *tree)
{
  // it's wrong to call CopyPartialBody if you don't have a starting point
  assert((*m_firstinvalid) != Branch::NONBRANCH);

  // we can't do a partial copy if the tree was changed too high up
  if ((*m_firstinvalid)->BranchGroup() == bgroupTip ||
      (*tree->m_firstinvalid)->BranchGroup() == bgroupTip) {
    m_timelist.CopyBody(tree->m_timelist,true);
  } else {
    m_timelist.CopyPartialBody(tree->m_timelist,tree->m_firstinvalid,
                               m_firstinvalid);
  }

  m_overallDL = tree->m_overallDL;
} 

//______________________________________________________________________________

TBranch_ptr Tree::CreateTip(const TipData& tipdata, const vector<LocusCell>& cells,
                         const vector<LocusCell>& movingcells)
{
  TBranch_ptr pTip = m_timelist.CreateTip(tipdata, cells, movingcells,
                                       *m_protorange);

  return pTip;
}

//_____________________________________________________________________________
void Tree::SetIndividualsWithTips(const vector<Individual>& indvec)
{
  m_individuals = indvec;
} /* SetIndividualsWithTips */

//_____________________________________________________________________________
void Tree::SetLocusVec(vector<Locus>* loc)
{
  m_pLocusVec = loc;
  m_totalsites = 0;
  vector<Locus>::const_iterator locit;
  for(locit = loc->begin(); locit != loc->end(); ++locit) {
     if (locit->GetSiteSpan().second > m_totalsites)
        m_totalsites = locit->GetSiteSpan().second;
  }
  delete m_protorange;
  m_protorange = new Range(m_totalsites, registry.GetForceSummary().GetSelectedSites());
} /* SetLocusVec */

//_____________________________________________________________________________
TreeSummary* Tree::SummarizeTree() const
{
  TreeSummary* treesum = registry.GetProtoTreeSummary().Clone();
  treesum->Summarize(*this);
  return treesum;
} /* SummarizeTree */

//_____________________________________________________________________________
double Tree::CalculateDataLikesForFixedLoci()
{
  unsigned long loc;
#ifndef NDEBUG
  vector<double> likes;
#endif
  double likelihood, overalllike(0);

  for (loc = 0; loc < m_pLocusVec->size(); ++loc) {
    const Locus& locus = (*m_pLocusVec)[loc];
    try {  // check for need to switch to normalization
      likelihood = locus.CalculateDataLikelihood(*this, false);
    }
    catch (datalikenorm_error& ex) { // normalization is set by thrower
      m_timelist.SetAllUpdateDLs();  // all subsequent loci will recalculate
                                   // the entire tree
      RunReport& runreport = registry.GetRunReport();
      runreport.ReportChat("\n", 0);
      runreport.ReportChat("Subtree of likelihood 0.0 found:  Turning on normalization and re-calculating.");

      likelihood = locus.CalculateDataLikelihood(*this, false);
    }
#ifndef NDEBUG
    likes.push_back(likelihood);
#endif
    overalllike += likelihood;   // add to accumulator

#if LIKETRACK
    ofstream of;
    of.open("likes1",ios::app);
    of << GetDLValue() << " ";
    of.close();
#endif
  }

#ifndef NDEBUG
    // This code tests a full likelihood recalculation against
    // the time-saving partial recalculation.  Slow but sure.
  double checkDL = 0.0;
  m_timelist.SetAllUpdateDLs();
  for (loc = 0; loc < m_pLocusVec->size(); ++loc) {
    const Locus& locus = (*m_pLocusVec)[loc];
    double likelihood = locus.CalculateDataLikelihood(*this, false);
    checkDL += likelihood;
    assert(fabs(likelihood - likes[loc])/likelihood < EPSILON);
  }
  assert(fabs(overalllike - checkDL)/overalllike < EPSILON);
#endif

  return overalllike;

} /* CalculateDataLikesForFixedLoci */

//____________________________________________________________________________

void Tree::SetupAliases(const vector<Locus>& loci)
{

// set up the aliases, this must be done after filling the tips with data!
  m_aliases.clear();
  vector<Locus>::const_iterator locus;
  for (locus = loci.begin(); locus != loci.end(); ++locus) {
     m_aliases.push_back(locus->GetDLCalc()->RecalculateAliases(*this,*locus));
  }

} /* SetupAliases */

//____________________________________________________________________________

vector<Branch_ptr> Tree::ActivateTips(Tree* othertree)
{
  vector<Branch_ptr> tips;
  m_firstinvalid = m_timelist.Begin();
  othertree->SetFirstInvalid();

  m_timelist.ClearBody();

  Branchiter brit;
  for (brit = m_timelist.FirstTip(); brit != m_timelist.End(); 
       brit = m_timelist.NextTip(brit))
  {
    tips.push_back(*brit);
  }

  return tips;

}
//_____________________________________________________________________________
Branch_ptr Tree::ActivateBranch(Tree* othertree)
{
  // NCuttable-1 because the root branch is of a cuttable type but is
  // not, in fact, a cuttable branch.

  // + 1 because it's helpful to count cuttable branches from 1, not 0.

  long rn = m_randomSource->Long(m_timelist.GetNCuttable()-1) + 1;

  Branchiter brit;
  Branch_ptr pActive = Branch::NONBRANCH;
  assert(m_timelist.IsValid());

  for (brit = m_timelist.Begin(); rn > 0; ++brit)
  {
    m_firstinvalid = brit;
    pActive = *brit;
    rn -= pActive->Cuttable();
  }
  
  Branchconstiter constfirstinvalid = m_firstinvalid;
  othertree->SetFirstInvalidFrom(constfirstinvalid);

  assert(pActive != Branch::NONBRANCH); // we didn't find any cuttable branches?

  Break(pActive);
  pActive->SetParent(0,Branch::NONBRANCH);
  pActive->SetParent(1,Branch::NONBRANCH);
  return pActive;
}

//______________________________________________________________________________

void Tree::SetFirstInvalidFrom(Branchconstiter& target)
{
  m_firstinvalid = m_timelist.FindEquivBranch(*target);
} /* SetFirstInvalidFrom */

//______________________________________________________________________________

Branch_ptr Tree::ActivateRoot()
{
  Branch_ptr pBase   = m_timelist.Base();  // Get the tree base and root.
  Branch_ptr pRoot   = pBase->Child(0);
  pBase->SetChild(0,Branch::NONBRANCH);             // Disconnect them.
  pRoot->SetParent(0,Branch::NONBRANCH);
  return pRoot;
}

//_____________________________________________________________________________

Branch_ptr Tree::ChoosePreferentiallyTowardsRoot(Tree* othertree)
{
  //Choose a random body branch, triangle weighted towards the root
  long nbranches = m_timelist.GetNBodyBranches();
  long weightedbranches = (nbranches * (nbranches + 1)) / 2;
  long chosenweight = m_randomSource->Long(weightedbranches) + 1;
  long chosenbranch(0L), weight(0L);
  while(weight < chosenweight) {
    ++chosenbranch;
    weight += chosenbranch;
  }

  long thisbranch(1L);
  for(m_firstinvalid = m_timelist.FirstBody();
      m_firstinvalid != m_timelist.End();
      m_firstinvalid = m_timelist.NextBody(m_firstinvalid), ++thisbranch) {
    if (thisbranch == chosenbranch) break;
  }
  assert(m_firstinvalid != m_timelist.End() && thisbranch == chosenbranch);

  // m_firstinvalid may be set incorrectly when there are multiple
  // branches with the same eventtime.
  m_firstinvalid = m_timelist.ResolveTiedTimes(m_firstinvalid);

  Branchconstiter constfirstinvalid = m_firstinvalid;
  othertree->SetFirstInvalidFrom(constfirstinvalid);

  return *m_firstinvalid;
}

//_____________________________________________________________________________
void Tree::AttachBase(Branch_ptr newroot)
{
  Branch_ptr pBase     = m_timelist.Base();     // Get the tree base.
  pBase->SetChild(0,newroot);            // Link the root to the base.
  newroot->SetParent(0,pBase);
}

//_____________________________________________________________________________

vector<Branch_ptr> Tree::FindAllBranchesAtTime(double eventT)
{
  Branch_ptr pBranch;
  vector<Branch_ptr> inactives;

  Branchiter brit;
  for (brit = m_timelist.Begin(); brit != m_timelist.End(); ++brit)
  {
    pBranch = *brit;
    if (eventT < pBranch->m_eventtime)
       break;

    // Sticks don't go in this list
    if (pBranch->Event() == btypeStick) continue; 

    if (pBranch->Parent(0))
    {
      if (eventT < pBranch->Parent(0)->m_eventtime)
         inactives.push_back(pBranch);
    }
    else assert(!pBranch->Parent(1));
  }

  return inactives;
}

//_____________________________________________________________________________

vector<Branch_ptr> Tree::FirstInterval(double eventT)
{
  return FindAllBranchesAtTime(eventT);
}

//_____________________________________________________________________________

vector<Branch_ptr> Tree::FindBranchesImmediatelyTipwardOf(Branchiter startbr)
{
  assert((*startbr)->m_eventtime != 0.0);
  Branchiter prevbr(startbr);
  --prevbr;  // assumes we have reverseable iterator
  while((*prevbr)->m_eventtime == (*startbr)->m_eventtime) {
    assert(prevbr != m_timelist.Begin());  // too far!
    --prevbr;
  }
  return FindAllBranchesAtTime((*prevbr)->m_eventtime);
}

//_____________________________________________________________________________
void Tree::NextInterval(Branch_ptr branch)
{
// deliberately blank; needed only in rectree
}

//____________________________________________________________________________

void Tree::GrowStumpAs(const Tree& src, bool nodlcells)
{
// deep copy the branchlist
   if (nodlcells) m_timelist.CopyWithoutDLCells(src.m_timelist);
   else m_timelist = src.m_timelist;

// now setup the tip pointers in the individuals correctly, since the
// individual copy ctor only makes empty containers of tips.
   unsigned long indiv;
   for(indiv = 0; indiv < m_individuals.size(); ++indiv) {
      StringVec1d tipnames = src.m_individuals[indiv].GetAllTipNames();
      vector<Branch_ptr> itips = GetTips(tipnames);
      m_individuals[indiv].SetTips(itips);
   }
}

//_____________________________________________________________________________
void Tree::Prune()
{
  m_timelist.Prune();
}

//_____________________________________________________________________________
Branch_ptr Tree::Coalesce(Branch_ptr child1, Branch_ptr child2, double tevent)
{
  Branch_ptr parent = Branch_ptr(new CBranch(*m_protorange));
  parent->m_eventtime = tevent;
  parent->CopyPartitionsFrom(child1);

  parent->SetChild(0,child1);
  parent->SetChild(1,child2);
  child1->SetParent(0,parent);
  child2->SetParent(0,parent);

  parent->SetDLCells(CollectCells());
  parent->SetUpdateDL();    // mark this branch for updating

  m_timelist.Collate(parent);

  return parent;

} /* Tree::Coalesce */

//_____________________________________________________________________________
Branch_ptr Tree::CoalesceActive(double eventT,
   Branch_ptr pActive1, Branch_ptr pActive2)
{

  // Create the parent of the two branches.
  Branch_ptr pParent = Coalesce(pActive1, pActive2, eventT);
  return pParent;
}

//_____________________________________________________________________________ 
Branch_ptr Tree::CoalesceInactive(double eventT,
   Branch_ptr pActive, Branch_ptr pInactive)
{
  // Create the parent of the two branches.
  Branch_ptr pParent = Branch_ptr(new CBranch (*m_protorange));
  pParent->m_eventtime  = eventT;
  pParent->CopyPartitionsFrom(pActive);

  // Hook children to parent
  pParent->SetChild(0,pInactive);
  pParent->SetChild(1,pActive);

  // Hook pInactive's old parent up to new parent
  pInactive->Parent(0)->ReplaceChild(pInactive, pParent);
  pInactive->SetParent(0,pParent);
  if (pInactive->Parent(1))
  {
    pInactive->Parent(1)->ReplaceChild(pInactive, pParent);
    pInactive->SetParent(1,Branch::NONBRANCH);
  }

  pActive->SetParent(0,pParent);

  pParent->SetDLCells(CollectCells());
  m_timelist.SetUpdateDLs(pParent);     // Set the update data likelihood flag.

  m_timelist.Collate(pParent);
  return pParent;
}

//_____________________________________________________________________________
Branch_ptr Tree::Migrate(double eventT, long pop, long maxEvents, Branch_ptr pActive)
{
  if (m_timelist.HowMany(btypeMig) >= maxEvents) {
    mig_overrun e;
    throw e;
  }

  Branch_ptr pParent = Branch_ptr(new MBranch (*m_protorange));
  pParent->m_eventtime  = eventT;
  pParent->CopyPartitionsFrom(pActive);
  pParent->SetPopulation(pop);

  pParent->SetChild(0,pActive);
  pActive->SetParent(0,pParent);

  m_timelist.Collate(pParent);
  return pParent;
}

//____________________________________________________________________________

Branch_ptr Tree::DiseaseMutate(double eventT, long endstatus, long maxevents,
   Branch_ptr pActive)
{

  if (m_timelist.HowMany(btypeDisease) >= maxevents) {
    dis_overrun e;
    throw e;
  }

  Branch_ptr pParent = Branch_ptr(new DBranch(*m_protorange));
  pParent->m_eventtime  = eventT;
  pParent->CopyPartitionsFrom(pActive);
  pParent->SetDisease(endstatus);

  pParent->SetChild(0,pActive);
  pActive->SetParent(0,pParent);

  m_timelist.Collate(pParent);
  return pParent;

} /* DiseaseMutate */

//____________________________________________________________________________

void Tree::SwapSiteDLs()
{
  Individual ind;

  // Pick an individual with phase-unknown sites at random.
  do {
    ind  = m_individuals[m_randomSource->Long(m_individuals.size())];
  } while (!ind.AnyPhaseUnknownSites());

  // Pick a phase marker at random.
  pair<long,long> locusmarker = ind.PickRandomPhaseMarker(*m_randomSource);
  long locus = locusmarker.first;
  long posn = locusmarker.second;

  // Pick two tips in the individual at random.
  vector<Branch_ptr> haps = ind.GetAllTips();
  long size = haps.size();
  long rnd1 = m_randomSource->Long(size);
  long rnd2 = m_randomSource->Long(size-1);
  if (rnd1 <= rnd2)
    rnd2++;

  Branch_ptr pTip1   = haps[rnd1];
  Branch_ptr pTip2   = haps[rnd2];

  // NB: Swap first DLCell only!

  assert(pTip1->GetNcells(0) == pTip2->GetNcells(0));

  // Get the data likelihood arrays.
  Cell_ptr dlcell1 = pTip1->GetDLCell(locus, 0, false);
  Cell_ptr dlcell2 = pTip2->GetDLCell(locus, 0, false);

  // Swap the cells at the marker position.
  dlcell1->SwapDLs(dlcell2,posn);

  // take care of needed data-likelihood corrections
  const Locus& loc = (*m_pLocusVec)[locus];
  m_aliases[locus] = loc.GetDLCalc()->RecalculateAliases(*this, loc);

  MarkForDLRecalc(pTip1);
  MarkForDLRecalc(pTip2);

} /* Tree::SwapSiteDL */

//____________________________________________________________________________

// PickNewSiteDLs is used when haplotype arranging for non-50/50 haplotype
//  probabilities.  Otherwise, SwapSiteDLs is used.
void Tree::PickNewSiteDLs()
{
  long ind;

  // Pick an individual with phase-unknown sites at random.
  do {
    ind  = m_randomSource->Long(m_individuals.size());
  } while (!m_individuals[ind].MultipleTraitHaplotypes());

  // Pick a locus/marker at random.
  pair<string, long> locusmarker = m_individuals[ind].PickRandomHaplotypeMarker();
  string lname = locusmarker.first;
  long marker  = locusmarker.second;

  // Have the individual in question pick up a new set of haplotypes.
  m_individuals[ind].ChooseNewHaplotypesFor(lname, marker);
  
  //Finish this routine in a subroutine, because chances are good that the
  // locus in question is a member of m_pMovingLocusVec, which we only
  // have if we're a RecTree.

  ReassignDLsFor(lname, marker, ind);
  return;
} /* PickNewSiteDLs */

void Tree::ReassignDLsFor(string lname, long marker, long ind)
{
  //Find out which locus we're dealing with.  We have its name.
  // If the locus in question is in m_pMovingLocusVec, we've already dealt with
  // it in RecTree::ReassignDLsFor
  long locus = FLAGLONG;
  for (unsigned long lnum=0; lnum<m_pLocusVec->size(); lnum++) {
    if ((*m_pLocusVec)[lnum].GetName() == lname) {
      locus = lnum;
    }
  }
  if (locus == FLAGLONG) {
    throw implementation_error("We have the segment name " + lname + ", but no segment matches it.  Did something happen to this tree?  To the segment in question?");
  }

  vector<Branch_ptr> haps = m_individuals[ind].GetAllTips();
  vector<LocusCell> cells = m_individuals[ind].GetLocusCellsFor(lname, marker);

  for (unsigned long tip=0; tip<haps.size(); tip++) {
    Cell_ptr origcell = haps[tip]->GetDLCell(locus, 0, false);
    Cell_ptr newcell  = cells[tip][0];
    origcell->SetSiteDLs(marker, newcell->GetSiteDLs(marker));
    MarkForDLRecalc(haps[tip]);
  }

  // take care of needed data-likelihood corrections
  const Locus& loc = (*m_pLocusVec)[locus];
  m_aliases[locus] = loc.GetDLCalc()->RecalculateAliases(*this, loc);

} /* Tree::ReassignDLsFor */

//___________________________________________________________________________

void Tree::MarkForDLRecalc(Branch_ptr markbr)
{

markbr->SetUpdateDL();
markbr->MarkParentsForDLCalc();

} /* Tree::MarkForDLRecalc */

//____________________________________________________________________________

rangevector Tree::GetLocusSubtrees(rangepair span) const
// this version is used to limit a multi-locus case to just the
// locus under consideration
{
  rangevector subtrees;
  subtrees.push_back(span); 
  return subtrees;
} /* GetLocusSubtrees */

//____________________________________________________________________________

vector<Branch_ptr> Tree::GetTips(StringVec1d& tipnames) const
{
vector<Branch_ptr> tips;
StringVec1d::iterator tname;
for(tname = tipnames.begin(); tname != tipnames.end(); ++tname)
   tips.push_back(GetTip(*tname));

return tips;

} /* GetTips */

//____________________________________________________________________________

Branch_ptr Tree::GetTip(const string& name) const
{
return m_timelist.GetTip(name);
} /* GetTip */

//____________________________________________________________________________

bool Tree::NoPhaseUnknownSites() const
{
IndVec::const_iterator ind;
unsigned long locus;
for (locus = 0; locus < m_pLocusVec->size(); ++locus) {
  for(ind = m_individuals.begin(); ind != m_individuals.end(); ++ind)
     if (ind->AnyPhaseUnknownSites()) return false;
}

return true;

} /* Tree:NoPhaseUnknownSites */

//____________________________________________________________________________

bool Tree::IsValid() const
{
  // NB:  This can only be called on a completed tree, not one
  // in mid-rearrangement.

  // each Branch validates itself
  Branchconstiter brit = m_timelist.Begin();
  long ncuttable = 0;
  for ( ; brit != m_timelist.End(); ++brit) {
    //LS TEST
    //(*brit)->PrintInfo();
    if (!(*brit)->CheckInvariant()) {
      return false;
    }
    ncuttable += (*brit)->Cuttable();
  }
  if (ncuttable != m_timelist.GetNCuttable()) {
    return false;
  }
  return m_timelist.IsValid();

//  if (FindContradictParts()) return false;
//  return true;
} /* IsValid */

//____________________________________________________________________________

bool Tree::operator==(const Tree& src) const
{
  if (m_pLocusVec->size() != src.m_pLocusVec->size()) return false;
  unsigned long i;
  for (i = 0; i < m_pLocusVec->size(); ++i) {
    if ((*m_pLocusVec)[i].GetNsites() != (*src.m_pLocusVec)[i].GetNsites()) return false;
  }
  if (m_overallDL != src.m_overallDL) return false;
  if (m_timelist != src.m_timelist) return false;

  // WARNING I don't guarantee these are the same tree, especially if
  // we are in mid-rearrangement, but they are at least very similar. --Mary
  return true;

} /* operator== */
 
//____________________________________________________________________________

bool Tree::SimulateDataIfNeeded()
{
  bool simulated = false;
  for (unsigned long loc=0; loc<m_pLocusVec->size(); loc++) {
    Locus& locus = (*m_pLocusVec)[loc];
    if (locus.GetShouldSimulate()) {
      simulated = true;
      locus.SimulateData(*this, m_totalsites);
    }
  }
  //Redo the aliases
  SetupAliases(*m_pLocusVec);
  //Note that we have to calculate all DLCells.
  m_timelist.SetAllUpdateDLs();
  return simulated;

}
//____________________________________________________________________________

void Tree::DLCheck(const Tree& other) const
{
cout << m_timelist.DLCheck(other.GetTimeList()) << endl;
} /* DLCheck */

//____________________________________________________________________________

void Tree::CreateStick(const ForceParameters& fp)
{

m_timelist.SetStickParameters(fp);

// create the first joint
JBranch_ptr br = JBranch_ptr(new JBranch(fp.GetRegionalThetas(),*m_protorange));
br->m_eventtime = 0.0;
m_timelist.CreateStick(br);

// make more joints as needed
m_timelist.ExtendStickToCoverTree();

} /* CreateStick */

//____________________________________________________________________________

void Tree::SetNewTimesFrom(Branchiter startpoint, const DoubleVec1d& newtimes)
{
assert(IsValid());
assert(!newtimes.empty());

bool updatefirstinvalid(startpoint == m_firstinvalid);

// first retime and remove all the branches from startpoint on...
vector<Branch_ptr> newbranches;
Branchiter branch(startpoint);
DoubleVec1d::const_iterator newtime(newtimes.begin());
for( ; branch != m_timelist.End(); ++newtime) {
   assert(newtime != newtimes.end());
// no special handling for recombinations needed here, assumed
// handled in calling code that sets up "newtimes".
   (*branch)->m_eventtime = *newtime;
   Branch_ptr newbr(*branch);
   branch = m_timelist.NextBody(branch);
   m_timelist.Remove(newbr);
   newbranches.push_back(newbr);
}

// now put them back in, setting datalikelihood update flags as needed
vector<Branch_ptr>::iterator newbranch;
for(newbranch = newbranches.begin(); newbranch != newbranches.end();
    ++newbranch) {
   Branchiter brit(m_timelist.Collate(*newbranch));
   if (newbranch == newbranches.begin() && updatefirstinvalid) {
      m_firstinvalid = brit;
   }
   m_timelist.SetUpdateDLs(*newbranch);
}

m_timelist.ExtendStickToCoverTree();

assert(IsValid());

} /* SetNewTimesFrom */

//____________________________________________________________________________

// Method to, if necessary, shrink interval lengths, usually at the root,
// in an attempt to prevent explosions in the next maximization.
// See the extensive comment in ChainManager::DoChain(), which is
// where this method is called.
//
bool Tree::GroomForGrowth(const DoubleVec1d& thetas,
		 const DoubleVec1d& growths,
		 double temperature) // temperature used only for debugging

{
  if (thetas.empty() || growths.empty() || thetas.size() != growths.size())
    throw implementation_error("Tree::GroomForGrowth() received an invalid theta and/or growth vector");

  bool treeWasModified = false;
  double starttime(0.), cumulativeShift(0.), prevIntervalLength(DBL_BIG);
  Branchiter brit;
  BranchBuffer lineages(registry.GetDataPack());
  for (brit = m_timelist.FirstTip(); brit != m_timelist.End();
    brit = m_timelist.NextTip(brit))
    lineages.UpdateBranchCounts((*brit)->m_partitions, true);

  // Iterate through the timelist, looking for fatal intervals.
  for (brit = m_timelist.FirstBody(); brit != m_timelist.End();
    brit = m_timelist.NextBody(brit))
    {
      Branch_ptr pBranch = *brit;
      unsigned long nchildren(0UL), worst_pop(0UL), npops = thetas.size();
      double smallestExpectedLength(DBL_BIG); // smallest E[dt] among the pops
      double expectedLength;
      LongVec1d xpartitions = lineages.GetBranchXParts();
      string msg;

      // Once we shrink an interval, we need to shift all intervals
      // below it by this amount.  The amount accumulates.
      if (cumulativeShift > 0.)
	{
	  pBranch->m_eventtime -= cumulativeShift;
	  pBranch->SetUpdateDL();
	}

      // Determine how many child branches we will need to update.
      switch(pBranch->Event()) {
      case btypeCoal:
	nchildren = static_cast<unsigned long>(NELEM);
        break;
      case btypeMig:
      case btypeDisease:
	nchildren = 1UL;
        break;
      case btypeRec:
	nchildren = 0UL; // not literally true; this is a "magic value"
        break;
      case btypeBase:
      case btypeTip:
      case btypeStick:
        // Unknown event type; we don't know how many
        // children or parents this event type has.
        assert(false);
        string msg = " Unrecognized branch type (\"" 
          +ToString(pBranch->Event())
          + "\") received by Tree::Groom().";
        throw implementation_error(msg);
        break;
      }
      double intervalLength = pBranch->m_eventtime - starttime;
      if (intervalLength <= 0.)
	{
	  if (0. == intervalLength && btypeRec == pBranch->Event() &&
	      prevIntervalLength > 0.)
	    {
	      // this is perfectly okay--rec branches come in pairs w/equal timestamps
	      lineages.UpdateBranchCounts(pBranch->m_partitions, true);
	      // Recombination:  The opposite of a coalescence--it has one child
	      // and two parents.  Each parent branch comes through separately
	      // and gets updated.  We need to avoid updating their shared child
	      // twice, so we choose to update the child when the "left" parent
	      // comes through.  (nchildren was set to the "magic value" of 0.)
	      if (pBranch->Child(0)->Parent(0) == pBranch)
		lineages.UpdateBranchCounts(pBranch->Child(0)->m_partitions, false);
	      starttime = pBranch->m_eventtime;
	      prevIntervalLength = intervalLength;
	      continue;
	    }
	  string msg = "Tree::GroomForGrowth(), encountered a time interval length of ";
	  msg += ToString(intervalLength) + ", followed by an event of type ";
	  msg += ToString(pBranch->Event()) + ".";
	  throw impossible_error(msg);
	}

      // Calculate and store E[dt] for each population.
      // Calculate E[dt] for coalescence + growth, regardless of
      // which event occurs at the bottom of the interval, because
      // coalescence + growth has the most extreme/sensitive values
      // for lnWait() and DlnWait().
      for (unsigned long pop = 0; pop < npops; pop++)
	{
	  const unsigned long& k = xpartitions[pop];
	  if (k > 1 && growths[pop] > 0.)
	    expectedLength =
	      (1./growths[pop])*ExpE1(SafeProductWithExp(k*(k-1)/(growths[pop]*thetas[pop]),
							   growths[pop]*starttime));
	  else
	    continue;
	  if (expectedLength < smallestExpectedLength)
	    {
	      smallestExpectedLength = expectedLength;
	      worst_pop = pop; // for debugging message only
	    }
	}

      // Note:  Calling an interval fatal if it's 100 times its new
      // expectation value is handwaving.  So far, 100 seems to be
      // a good sanity check, but this value might need to be reduced
      // to 20 or 10 if growth is very large.
      if (smallestExpectedLength < DBL_BIG &&
	  intervalLength >= 100. * smallestExpectedLength)
	{
	  msg = "\nTemperature " + ToString(temperature) + ", population "
	    + ToString(worst_pop) + ", k = " + ToString(xpartitions[worst_pop])
	    + ", interval length is " + ToString(intervalLength)
	    + ", expected value is " + ToString(smallestExpectedLength)
	    + ".  Interval ends with a " + ToString(pBranch->Event())
	    + " event.  Time stamps are " + ToString(starttime) + " and "
	    + ToString(pBranch->m_eventtime) + "; theta = " + ToString(thetas[worst_pop])
            + " and g = " + ToString(growths[worst_pop]) + ".  ";
	  
	  // Shrink this overlong interval to a value that is between
	  // 1/2 and 3/2 times the expectation value under the new parameter values
	  // for the "worst" population.
	  double factor = 0.5 + m_randomSource->Float();
	  double newIntervalLength = factor * smallestExpectedLength;

	  if (newIntervalLength < starttime * numeric_limits<double>::epsilon())
	    {
	      msg += "Tried to shrink this interval to a length of "
		+ ToString(newIntervalLength) + ", but this is smaller than the "
		+ "minimum length of "
		+ ToString(starttime * numeric_limits<double>::epsilon())
		+ " that can be added to " + ToString(starttime)
		+ " without being lost to rounding error.  Giving up and "
		+ "copying the cold tree into the tree for temperature "
		+ ToString(temperature) + "....\n";
	      registry.GetRunReport().ReportDebug(msg);
	      return false;
	    }

	  pBranch->m_eventtime = starttime + newIntervalLength;
	  pBranch->SetUpdateDL();
	  treeWasModified = true;
	  
	  msg += "Shrinking this interval to a length of "
	    + ToString(newIntervalLength) + ", with a new ending time "
	    + "stamp of " + ToString(pBranch->m_eventtime) + ".\n";
	  registry.GetRunReport().ReportDebug(msg);

	  // Once we shrink an interval, we need to shift all intervals
	  // below it to reflect this.
	  cumulativeShift += intervalLength - newIntervalLength;

	}

      // Prepare for next loop iteration (next branch in the timelist).
      for (unsigned long i = 0; i < nchildren; i++)
	lineages.UpdateBranchCounts(pBranch->Child(i)->m_partitions,
           false);
      lineages.UpdateBranchCounts(pBranch->m_partitions, true);

      // Recombination:  The opposite of a coalescence--it has one child
      // and two parents.  Each parent branch comes through separately
      // and gets updated.  We need to avoid updating their shared child
      // twice, so we choose to update the child when the "left" parent
      // comes through.  (nchildren was set to the "magic value" of 0.)
      if (btypeRec == pBranch->Event() &&
	  pBranch->Child(0)->Parent(0) == pBranch)
	lineages.UpdateBranchCounts(pBranch->Child(0)->m_partitions,
           false);

      starttime = pBranch->m_eventtime;
      prevIntervalLength = intervalLength; // used to detect successive zero-lengths, if any
    }

  if (treeWasModified)
    CalculateDataLikes();

  return true;
}

//____________________________________________________________________________


// Method to, if necessary, shrink interval lengths, usually at the root,
// in an attempt to prevent explosions in the next maximization.
// See the extensive comment in ChainManager::DoChain(), which is
// where this method is called.
//
bool Tree::GroomForLogisticSelection(const DoubleVec1d& thetas,
		 double s, // logistic selection coefficient
		 double temperature) // temperature used only for debugging

{
  if (0. == s)
    return true;
  if (2 != thetas.size() || 0. == thetas[0] || 0. == thetas[1])
    {
      string msg = "Tree:GroomForLogisticSelection(), received invalid Theta vector.";
      throw implementation_error(msg);
    }

  bool treeWasModified = false;
  double starttime(0.), cumulativeShift(0.), shrinkageFactor(1.);
  double max_allowed_starttime(DBL_BIG);
  Branchiter brit, firstBadInterval, lastBadInterval;
  string msg;

  if (s > 0.)
    max_allowed_starttime = (EXPMAX - log(thetas[1]))/s;
  else
    max_allowed_starttime = (EXPMAX - log(thetas[0]))/(-s);

  firstBadInterval = lastBadInterval = m_timelist.End();

  // Iterate through the timelist, looking for fatal intervals.
  for (brit = m_timelist.FirstBody(); brit != m_timelist.End();
    brit = m_timelist.NextBody(brit))
    {
      if ((*brit)->m_eventtime >= max_allowed_starttime)
	{
	  Branchiter brit2 = brit;
	  firstBadInterval = brit;
	  while (brit2 != m_timelist.End())
	    {
	      lastBadInterval = brit2;
	      brit2 = m_timelist.NextBody(brit2);
	    }
	  break;
	}
      starttime = (*brit)->m_eventtime;
    }

  if (firstBadInterval == m_timelist.End())
    return true; // nothing to shrink

  if ((*lastBadInterval)->m_eventtime - starttime <= 0.)
    {
      msg += "Tree::GroomForLogisticSelection(), encountered an interval ";
      msg += "(or sum of intervals) of length ";
      msg += ToString((*lastBadInterval)->m_eventtime - starttime) + ".";
      throw implementation_error(msg);
    }

  shrinkageFactor = (max_allowed_starttime - starttime) /
    ((*lastBadInterval)->m_eventtime - starttime);

  if (shrinkageFactor <= 0.)
    {
      msg += "Tree::GroomForLogisticSelection(), computed an invalid shrinkage ";
      msg += "factor of " + ToString(shrinkageFactor) + ".";
      throw implementation_error(msg);
    }

  // If there are any fatal intervals, iterate through these,
  // and shrink them.
  for (brit = firstBadInterval; brit != m_timelist.End();
    brit = m_timelist.NextBody(brit))
    {
      Branch_ptr pBranch = *brit;

      // Once we shrink an interval, we need to shift all intervals
      // below it by this amount.  The amount accumulates.
      if (cumulativeShift > 0.)
	{
	  pBranch->m_eventtime -= cumulativeShift;
	  pBranch->SetUpdateDL();
	}

      double intervalLength = pBranch->m_eventtime - starttime;
      pBranch->m_eventtime = starttime + intervalLength*shrinkageFactor;
      pBranch->SetUpdateDL();
      treeWasModified = true;
      
      // Once we shrink an interval, we need to shift all intervals
      // below it to reflect this.
      cumulativeShift += intervalLength - intervalLength*shrinkageFactor;
      
      starttime = pBranch->m_eventtime;
    }

  if (treeWasModified)
    CalculateDataLikes();

  return true;
}

//____________________________________________________________________________
//____________________________________________________________________________

Tree* PlainTree::Clone() const
{
  return new PlainTree(*this,false);

} /* PlainTree::Clone */

//____________________________________________________________________________

Tree* PlainTree::MakeStump() const
{
  return new PlainTree(*this,true);
} /* PlainTree stump maker */


//____________________________________________________________________________


void PlainTree::CalculateDataLikes()
{
  m_overallDL = CalculateDataLikesForFixedLoci();
  m_timelist.ClearUpdateDLs();         // reset the updating flags
}

//____________________________________________________________________________

void PlainTree::Break(Branch_ptr pBranch)
{
  Branch_ptr pParent = pBranch->Parent(0);

  if (pParent->CanRemove(pBranch))
  {
    Break(pParent);                                  // Recursion
    m_timelist.Remove(pParent);

    pParent = pBranch->Parent(1);
    if (pParent)
    {
      Break(pParent);                                // Recursion
      m_timelist.Remove(pParent);
    }
  }
}

