//$Id: tree.h,v 1.56 2007/09/28 22:00:13 lpsmith Exp $

#ifndef TREE_H
#define TREE_H

/* 
 Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

 This software is distributed free of charge for non-commercial use
 and is copyrighted.  Of course, we do not guarantee that the software
 works, and are not responsible for any damage you may cause or have.
*/

/*******************************************************************
 Class Tree represents a genealogy (not necessarily a "tree" in the
 recombinant cases).  It has two subtypes, a PlainTree (no 
 recombination) and a RecTree (with recombination) because the
 recombination machinery is too expensive to carry if you don't need
 it.

 Normally all Trees used during rearrangement come from either
 copying an existing tree or copying the prototype tree in Registry.

 DEBUG:  This class is a monster.  Anything to make it simpler
 would be good.

 Written by Jim Sloan, heavily revised by Mary Kuhner

 02/08/05 Pulled out locus specific stuff into LocusLike class -- Mary
 2004/9/15 Killed the LocusLike class, merging into Locus
 2005/3/1 Moved alias info, reluctantly, from DLCalc to Tree (because
    DLCalc is a shared object and the alias is not shareable!) -- Mary
 2006/02/13 Begun to add jointed stick
*******************************************************************/

#include <math.h>
#include "timelist.h"  // for TimeList member
#include "individual.h"  // for IndVec member
#include "locus.h"
#include <vector>
#include "vectorx.h"
#include "constants.h"
#include "definitions.h"
#include "rangex.h"

// typedef also used by Event and Arranger
typedef std::pair<Branch_ptr, Branch_ptr> branchpair; 

class TreeSummary;  // return type of SummarizeTree()
class TipData;
class Random;
class StickForce;
class ForceParameters;

class Tree
{
private:
         Tree(const Tree&);              // undefined
  Tree&  operator=(const Tree&);         // undefined

  Branchiter    m_firstinvalid;          // a "timelist" iterator to the
                                         // the branch after the cut during
					 // rearrangement.  Set by
					 // ActivateBranch(). 
					 // Used only mid-rearrangement
					 // so that CopyPartialBody() can
					 // be used for speed.
					 // This member is not copied!
  vector<LocusCell> m_protoCells;        // cache of locuscells for all loci

  vector<Branch_ptr> FindAllBranchesAtTime(double eventT);

protected:
                Tree();
                Tree(const Tree& tree, bool makestump);
  IndVec        m_individuals;           // associates tips from same individual
  double        m_overallDL;            // data log-likelihood
  LongVec2d     m_aliases;              // dim: loci x markers
  Range*        m_protorange;           // prototypical range of active sites 
  Random*       m_randomSource;         // non-owning pointer
  TimeList      m_timelist;             // all lineages
  vector<Locus>* m_pLocusVec;           // not owning.  Also, not const
  //                                        because of simulation
  long          m_totalsites;           // span of entire region, including gaps

  // Protected construction primitive
          void    GrowStumpAs(const Tree& src, bool nodlcells);

  // Protected rearrangement primitives
  virtual void    Break(Branch_ptr pBranch) = 0;
          void    SetFirstInvalidFrom(Branchconstiter& target);
          void    SetFirstInvalid() {m_firstinvalid = m_timelist.Begin();};
          void    MarkForDLRecalc(Branch_ptr br);

  // Protected Branch-management primitives
  const vector<LocusCell>& CollectCells();  // get locuscells for all loci
          vector<Branch_ptr>  GetTips(StringVec1d& names)  const;

public:

  // Creation and destruction
  virtual           ~Tree();
  virtual Tree     *Clone()                 const = 0;
          Tree     *CloneWithoutDLCells()   const;
  virtual Tree     *MakeStump()             const = 0;
  virtual void      Clear();
  virtual void      CopyTips(const Tree* tree);
  virtual void      CopyBody(const Tree* tree);
  virtual void      CopyStick(const Tree* tree);
  virtual void      CopyPartialBody(const Tree* tree);
          TBranch_ptr CreateTip(const TipData& tipdata,
                              const vector<LocusCell>& cells,
                              const vector<LocusCell>& movingcells);

  // Getters
          TimeList& GetTimeList()                 { return m_timelist; };
    const TimeList& GetTimeList()           const { return m_timelist; };
  virtual rangevector GetLocusSubtrees(rangepair span) const;
          Branch_ptr     GetTip(const string& name) const;
          double      RootTime()                       const 
            {return m_timelist.RootTime();};
          bool        NoPhaseUnknownSites()            const;

	  bool GroomForGrowth(const DoubleVec1d& thetas,
                     const DoubleVec1d& growths,
                     double temperature);
	  bool GroomForLogisticSelection(const DoubleVec1d& thetas,
		     double s, // logistic selection coefficient or zero
                     double temperature);

  // Setters
          void      SetIndividualsWithTips(const vector<Individual>& indvec);
          void      SetLocusVec(vector<Locus>* loc);

  // Likelihood manipulation
  virtual void      CalculateDataLikes() = 0;
  virtual double    CalculateDataLikesForFixedLoci();
          double    GetDLValue()            const { return m_overallDL; };
    const LongVec1d& GetAliases(long loc) const { return m_aliases[loc]; };
          void      SetupAliases(const std::vector<Locus>& loci);

  // Rearrangement primitives
  virtual vector<Branch_ptr> ActivateTips(Tree* othertree);
  virtual Branch_ptr   ActivateBranch(Tree* othertree);
  virtual Branch_ptr   ActivateRoot();
  // ChoosePreferentiallyTowardsRoot does not break the tree, but does
  // set m_firstinvalid to the interval containing the return value
          Branch_ptr   ChoosePreferentiallyTowardsRoot(Tree* othertree);
  virtual void      AttachBase(Branch_ptr newroot); 
  virtual vector<Branch_ptr> FindBranchesImmediatelyTipwardOf(Branchiter start);
  virtual vector<Branch_ptr> FirstInterval(double eventT);
  virtual void      NextInterval(Branch_ptr branch );
  virtual void      Prune();
          void      SwapSiteDLs();
  virtual void      PickNewSiteDLs();
  virtual void      ReassignDLsFor(string lname, long marker, long ind);
	  void      CreateStick(const ForceParameters& fp);
	  void      SetNewTimesFrom(Branchiter start, 
                       const DoubleVec1d& newtimes);
	  // FindStickAtTime throws an implementation error on failure
          JBranch_ptr FindStickAtTime(double time) const
                         {return m_timelist.FindStickAtTime(time);};

  // Force-specific rearrangement primitives
          Branch_ptr   Coalesce(Branch_ptr child1, Branch_ptr child2, double tevent); 

  virtual Branch_ptr   CoalesceActive(double eventT, Branch_ptr active1, Branch_ptr active2);
  virtual Branch_ptr   CoalesceInactive(double eventT, Branch_ptr active, Branch_ptr inactive);
  virtual Branch_ptr   Migrate(double eventT, long topop, long maxEvents,
                      Branch_ptr active);
  virtual Branch_ptr   DiseaseMutate (double eventT, long endstatus,
                      long maxEvents, Branch_ptr active);

  // TreeSummaryFactory;
  TreeSummary* SummarizeTree() const;

  // Invariant checking
          bool      IsValid()          const;   // check invariants
          bool      operator==(const Tree& src) const;  // compare trees
          bool      operator!=(const Tree& src) const {return !(*this == src); };

  // TimeList call throughs (we provide a public front for TimeList
  void SetStickParams(const ForceParameters& fp) 
                     { m_timelist.SetStickParameters(fp); }; 

  //Simulation
  virtual bool  SimulateDataIfNeeded();
  virtual long  NumberOfRecombinations() = 0;

  // TreeSizeArranger
  virtual void  SetActiveLinksFrom(const BranchBuffer&) {/* no-op */;};
  virtual void  ClearActiveLinks() { }; // no-op

  //Final Coalescent code (no-ops for PlainTree)
  virtual void SetFCFrom(BranchBuffer& actives, BranchBuffer& inactives) = 0;
  
  // Debugging code
  // MakeCoalescent strips all migration nodes from the tree and
  // forces all remaining coalescences to their expectation times.
  // DO NOT use with migration!
    void MakeCoalescent(double theta) {m_timelist.MakeCoalescent(theta);};

  // DLCheck returns an error message with the first marker each
  // branch pair differs at, unmentioned branches don't differ.
    void DLCheck(const Tree& other) const;

};

//______________________________________________________________________________
//______________________________________________________________________________

class PlainTree : public Tree
{
  public:

    // Creation and destruction
                      PlainTree() : Tree()  {};
                      PlainTree(const Tree& tree, bool makestump) :
                        Tree(tree,makestump)       {};
    virtual          ~PlainTree()        {};
    virtual Tree     *Clone()                 const;
    virtual Tree     *MakeStump()             const;
    virtual void      CalculateDataLikes();
    virtual void      Break(Branch_ptr pBranch);
    virtual long      NumberOfRecombinations() {return 0L;};

    //Final Coalescent code (no-ops for PlainTree)
    virtual void SetFCFrom(BranchBuffer& actives, BranchBuffer& inactives) {};
    // Yes, everything else is the same as Tree.

};

//______________________________________________________________________________
//______________________________________________________________________________

//To keep track of final coalescences:
typedef std::map<long, rangeset> rangesetcount;
typedef std::map<long, rangeset>::iterator RSCIter;
typedef std::map<long, rangeset>::const_iterator RSCcIter;

class RecTree : public Tree
{
private:

    long      m_activeLinks;
    long      m_openLinks;

    rangesetcount m_currentFC;
    rangeset  m_newactive; //For debugging only.

    void      UpdateRange(Branch_ptr pBranch, bool dofc);
    void      RecurseBreak(Branch_ptr pBranch);
    std::set<long> GetBreakpoints() const;

protected:
    vector<Locus>*    m_pMovingLocusVec;// not owning.  But we can change 'em!
    vector<LocusCell> m_protoMovingCells;

    virtual  void     Break(Branch_ptr branch);
    const vector<LocusCell>& CollectMovingCells();

public:
    // Creation and destruction
                      RecTree();
                      RecTree(const RecTree& tree, bool makestump);
                      ~RecTree()                        {};

    virtual  Tree    *Clone()     const;
    virtual  Tree    *MakeStump() const;
    virtual  void     Clear();
    virtual  void     CopyTips(const Tree* tree);
    virtual  void     CopyBody(const Tree* tree);
    virtual  void     CopyPartialBody(const Tree* tree);

    // Getters
    virtual  rangevector GetLocusSubtrees(rangepair span) const;
             long     ActiveLinks() const           { return m_activeLinks; };
             long     OpenLinks() const             { return m_openLinks; };
    unsigned long GetNumMovingLoci() {return m_pMovingLocusVec->size();};
    virtual  bool DoesThisLocusJump(long mloc) const;
    virtual  bool AnyRelativeHaplotypes() const;

    // Setters
    void      SetMovingLocusVec(vector<Locus>* locs);
    void      SetMovingMapposition(long mloc, long site);

    // Likelihood manipulation
    virtual void        CalculateDataLikes();
    virtual double      CalculateDataLikesForMovingLocus(long mloc);
    virtual DoubleVec1d CalculateDataLikesForFloatingLocus(long mloc);
    virtual DoubleVec1d CalculateDataLikesWithRandomHaplotypesForFloatingLocus(long mloc);
    virtual void CalculateDataLikesForAllHaplotypesForFloatingLocus(long mloc,
                                                          DoubleVec1d& mlikes);
    virtual bool UpdateDataLikesForIndividualsFrom(long ind, long mloc,
                                                   DoubleVec1d& mlikes);

    // Rearrangement primitives
    virtual  vector<Branch_ptr> ActivateTips(Tree* othertree);
    virtual  Branch_ptr  ActivateBranch(Tree* othertree);
    virtual  Branch_ptr  ActivateRoot();
    virtual  void     AttachBase(Branch_ptr newroot);
    virtual  vector<Branch_ptr> FirstInterval(double eventT);
    virtual  void     NextInterval(Branch_ptr branch);
    virtual  void     Prune();
    virtual  void     ReassignDLsFor(string lname, long marker, long ind);

    // Force-specific rearrangement primitives
    virtual  Branch_ptr  CoalesceActive(double eventT, Branch_ptr active1, Branch_ptr active2);
    virtual  Branch_ptr  CoalesceInactive(double eventT, Branch_ptr active, Branch_ptr inactive);
    virtual  Branch_ptr  Migrate(double eventT, long topop, long maxEvents,
                        Branch_ptr active);
    virtual  Branch_ptr  DiseaseMutate (double eventT, long endstatus,
                        long maxEvents, Branch_ptr active);
             branchpair RecombineActive(double eventT, long maxEvents, 
                        FPartMap fparts, Branch_ptr active, long recsite);
             branchpair RecombineInactive(double eventT, long maxEvents, 
                        FPartMap fparts, Branch_ptr active, long recsite);

  //Map Summary
  DoubleVec2d  GetMapSummary();
  std::set<long> IgnoreDisallowedSubTrees(std::set<long> breakpoints,
                                          rangeset allowedranges);
  DoubleVec1d ZeroDisallowedSites(DoubleVec1d datalikes,
                                  rangeset allowedranges);
  void RandomizeMovingHaplotypes(long mlocus);

  //Simulation
  virtual bool SimulateDataIfNeeded();
  virtual long NumberOfRecombinations();

  // TreeSizeArranger
  virtual void  SetActiveLinksFrom(const BranchBuffer& brbuffer);
  virtual void  SetOpenLinksFrom(const BranchBuffer& brbuffer); //debugging
  virtual void  ClearActiveLinks() { m_activeLinks = 0; };

  //Final Coalescent code (no-ops for PlainTree)
  virtual void SetFCFrom(BranchBuffer& actives, BranchBuffer& inactives);
  virtual void IncreaseFCCountOf(rangeset rs);
  virtual void DecreaseFCCountOf(rangeset rs);

  virtual void CheckFC(const rangeset& newlycoalesced, Branch_ptr branch);
  virtual void UpdateFCToIncludeFrom(const rangeset& newfc, Branch_ptr branch);
  virtual void UpdateTimeListToExcludeFromFC(Branch_ptr branch);

  //Debugging
  virtual bool FCMatchesList(BranchBuffer& actives, BranchBuffer& inactives);
  virtual void PrintTipData(long mloc, long marker);
  virtual void PrintFC();
  virtual void PrintRangeSetCount(const rangesetcount& rsc);
  virtual rangesetcount RemoveEmpty(const rangesetcount& rsc);
};

#endif /* TREE_H */
