//$id: treesum.cpp,v 1.23 2002/09/10 23:10:09 ewalkup Exp $  */  */  */

/* 
   Copyright 2002  Peter Beerli, Mary Kuhner, Jon Yamato and Joseph Felsenstein

   This software is distributed free of charge for non-commercial use
   and is copyrighted.  Of course, we do not guarantee that the software
   works, and are not responsible for any damage you may cause or have.
*/
#include "treesum.h"
#include "tree.h"
#include "summary.h"
#include <map>
#include <iostream> 
#include <fstream>
#include <sstream>
#include "registry.h"
#include "constants.h"
#include <iterator>
#include <string>
#include "chainmanager.h"
#include "xmlsum_strings.h" // for xml sumfile strings
#include "force.h" // for setting up the recsbypartition vector of intervals
                   // in a RecTreeSummary

#ifdef DMALLOC_FUNC_CHECK
#include "/usr/local/include/dmalloc.h"
#endif

using namespace std;

typedef std::map<string, double> Forcemap;
typedef std::map<string, double>::iterator Forcemapiter;

//____________________________________________________
//____________________________________________________
//____________________________________________________

// This copy constructor produces a usable EMPTY
// r++ comment'"
// No method is provided to copy the contents of a
// TreeSummary; it is considered too expensive.

TreeSummary::TreeSummary(const TreeSummary& src) 
  : intervalData(),
    nCopies(0)
{
  Sumconst_iter it = src.summaries.begin();
  for ( ; it != src.summaries.end(); ++it) {
    Summary* sum = (*it).second->Clone(intervalData);
    summaries.insert(make_pair((*it).first, sum));
  }
  CacheSummaries();
} /* Copy ctor */

//____________________________________________________

TreeSummary* TreeSummary::Clone() const
{
  return new TreeSummary(*this);
} /* Clone */

//____________________________________________________

TreeSummary::~TreeSummary()
{
  Sumiter it = summaries.begin();
  for ( ; it != summaries.end(); ++it) {
    delete (*it).second;
  }
} /* dtor */
//____________________________________________________

void TreeSummary::Compress()
{
  Summap::iterator it = summaries.begin();
  bool allshort = true;

  for ( ; it != summaries.end(); ++it) {
    allshort = (*it).second->Compress() && allshort;
  }
  if (allshort) intervalData.clear();

} /* Compress */

//____________________________________________________

// MDEBUG Currently the tree is summarized if either the tree or
// the stick has changed.  This may be inefficient if tree changes
// are rare and stick changes are common.  Revisit this decision
// if maximization slows down.
void TreeSummary::Summarize(const Tree& tree)
{

#ifdef  STATIONARIES
  ofstream of;
  of.open(INTERVALFILE.c_str(),ios::app);
#endif

  nCopies = 1;

  const TimeList& timeList = tree.GetTimeList();

  // Tally up populations of tips into lineages vector

  BranchBuffer lineages(registry.GetDataPack());
  Branchconstiter brit;

  for (brit = timeList.FirstTip(); brit != timeList.End(); 
    brit = timeList.NextTip(brit))
    lineages.UpdateBranchCounts((*brit)->m_partitions);

  // Traverse body of tree collecting information

#ifdef  STATIONARIES
  long interval = 0;
  double topTime = 0.0;
#endif

  for (brit = timeList.FirstBody(); brit != timeList.End();
    brit = timeList.NextBody(brit))
    {
      // score the branch
      Branch_ptr pBranch = *brit;

#ifdef  STATIONARIES
      double bottomTime = pBranch->m_eventtime;
      of << "int" << interval << " " << bottomTime - topTime << endl;
      ++interval;
      topTime = bottomTime;
#endif

      // This call adjusts lineages as a side effect

      pBranch->ScoreEvent(*this, lineages);
    }

  // now summarize the stick
  if (!(timeList.NoStick())) {
     timeList.ScoreStick(*this);   
  }

  Compress();
  CacheSummaries();

#ifdef  STATIONARIES
  of.close();
#endif

} /* Summarize */

//____________________________________________________

#ifdef STATIONARIES
void TreeSummary::DumpStationariesData(const Tree& tree) const {
  ofstream of;
  of.open(INTERVALFILE.c_str(),ios::app);

  const TimeList& timeList = tree.GetTimeList();
  Branchconstiter brit;

  long interval = 0;
  double topTime = 0.0;

  for (brit = timeList.FirstBody(); brit != timeList.End();
    brit = timeList.NextBody(brit))
    {
      Branch_ptr pBranch = *brit;

      double bottomTime = pBranch->m_eventtime;
      of << "int" << interval << " " << bottomTime - topTime << endl;
      ++interval;
      topTime = bottomTime;

    }

  of.close();

  of.open(MIGFILE.c_str(),ios::app);
  long migs = timeList.HowMany(btypeMig);
  of << migs << endl;
  of.close();

  of.open(DISFILE.c_str(),ios::app);
  long ndisevents = timeList.HowMany(btypeDisease);
  of << ndisevents << endl;
  of.close();

  of.open(RECFILE.c_str(),ios::app);
  long recs = timeList.HowMany(btypeRec);
  // divide by two because HowMany gives the number
  // of branches, not events
  recs = recs / 2;
  of << recs << endl;
  of.close();

} /* DumpStationariesData */
#endif

//____________________________________________________
// This is OPTIMIZATION code which supports the
// specialized GetMigSummary and so forth functions.
// It's ugly and force-specific but a significant
// speedup over calling GetSummary(lamarcstrings::MIG).
 
void TreeSummary::CacheSummaries()
{
  coalsummary = GetSummary(force_COAL);
  migsummary = GetSummary(force_MIG);
  diseasesummary = GetSummary(force_DISEASE);
  recsummary = GetSummary(force_REC);
  growsummary = GetSummary(force_GROW);

} /* CacheSummaries */

//____________________________________________________

map<force_type, DoubleVec1d> TreeSummary::InspectSummary() const
{
  // for each force, get a vector containing the number of events
  // for each parameter; return them in a map indexed by force

  Summap::const_iterator summary;
  map<force_type, DoubleVec1d> counts;

  for (summary = summaries.begin(); summary != summaries.end(); ++summary) {
    counts.insert(make_pair(summary->first, summary->second->GetShortPoint()));
  }

  return counts;
} /* InspectSummary */

//____________________________________________________

void TreeSummary::AdjustSummary(const map<force_type, DoubleVec1d>& counts, long region)
{

  map<force_type, DoubleVec1d>::const_iterator fit;

  for (fit = counts.begin(); fit != counts.end(); ++fit) {
    Summary* sum = GetSummary((*fit).first);
    assert(sum);  // no summary?!
    sum->AdjustSummary((*fit).second, region);
  }

} /* AdjustSummary */

//____________________________________________________

Summary* TreeSummary::GetSummary(force_type type)
{

  // Returning NULL is not necessarily indicative of an error
  // state--some forces, notably Growth, do not have corresponding
  // Summaries.

  Sumiter it = summaries.find(type);
  if (it == summaries.end()) return NULL;
  else return (*it).second;

} /* GetSummary */

//____________________________________________________

Summary const* TreeSummary::GetSummary(force_type type) const
{

  // Returning NULL is not necessarily indicative of an error
  // state--some forces, notably Growth, do not have corresponding
  // Summaries.

  Sumconst_iter it = summaries.find(type);
  if (it == summaries.end()) return NULL;
  else return (*it).second;

} /* GetSummary const */

//____________________________________________________________

void TreeSummary::AddSummary(force_type type, Summary* sum)
{
  // we should not be adding an already present type!
  assert(summaries.find(type) == summaries.end());

  summaries.insert(make_pair(type, sum));

} /* AddSummary */

//____________________________________________________
//____________________________________________________

RecTreeSummary::RecTreeSummary() : TreeSummary()
{
  const ForceVec partforces(registry.GetForceSummary().GetPartitionForces());
  ForceVec::const_iterator pforce;
  for(pforce = partforces.begin(); pforce != partforces.end(); ++pforce) {
     vector<Interval*> sums;
     vector<vector<Interval*> > recs((*pforce)->GetNPartitions(),sums);
     m_recsbypart.push_back(recs);
  }
} /* RecTreeSummary::ctor */

//____________________________________________________

RecTreeSummary::RecTreeSummary(const RecTreeSummary& src) : TreeSummary(src),
m_recsbypart(src.m_recsbypart)
{
 
} /* RecTreeSummary::copy ctor */

//____________________________________________________

TreeSummary* RecTreeSummary::Clone() const
{
  return new RecTreeSummary(*this);
} /* Clone */

//____________________________________________________

void RecTreeSummary::Summarize(const Tree& tree)
{
#ifdef  STATIONARIES
  ofstream of;
  of.open(INTERVALFILE.c_str(), ios::app);
#endif

  nCopies = 1;
  BranchBuffer lineages(registry.GetDataPack());

  const TimeList& timeList = tree.GetTimeList();

  // Tally up populations of tips into lineages vector

  Branchconstiter brit;
  long activelinks = 0;

  for (brit = timeList.FirstTip(); brit != timeList.End();
    brit = timeList.NextTip(brit))
    {
      lineages.UpdateBranchCounts((*brit)->m_partitions);
      activelinks += (*brit)->m_range.ActiveLinks();
    }

  // Traverse tree collecting information

#ifdef  STATIONARIES
  long interval = 0;
  double topTime = 0.0;
  double bottomTime = 0.0;
#endif

  for (brit = timeList.FirstBody(); brit != timeList.End();
    brit = timeList.NextBody(brit))
    {
      Branch_ptr pBranch = *brit;
      pBranch->ScoreEvent(*this, lineages, activelinks);

#ifdef  STATIONARIES
      bottomTime = pBranch->m_eventtime;
      of << "int" << interval << " " << bottomTime - topTime << endl;
      ++interval;
      topTime = bottomTime;
      // we assume that paired recombination branches are adjacent in
      // the timelist.
      if (pBranch->Event() == btypeRec) {
         dynamic_cast<RecSummary*>(recsummary)->
	    AddToRecombinationCounts(pBranch->m_partitions);
         AddRecToRecsByPart(pBranch->m_partitions,recsummary->GetLastAdded());
         brit = timeList.NextBody(brit);
      }
#endif
    }

  Compress();

} /* Summarize */

//____________________________________________________

void RecTreeSummary::AddRecToRecsByPart(const LongVec1d& membership, 
  Interval* newrec)
{
  LongVec1d::size_type partforce;
  for(partforce = 0; partforce < membership.size(); ++partforce) {
     m_recsbypart[partforce][membership[partforce]].push_back(newrec);
  }
} /* SetupRecByPartitionPointers */

//____________________________________________________

const vector<vector<vector<Interval*> > >&
RecTreeSummary::GetRecsByPart() const
{
  return m_recsbypart;
} /* GetRecsByPart */

//____________________________________________________
// function: writes tree summaries to sumfile
// notes:  doesn't assume cached summary ptrs set to null for forces not
//   in use/existence since cached summary ptrs could look valid for a
//   force when the force isn't even on.  Should test for this condition if
//   going to optimize out the map lookup (call to GetSummary)

void TreeSummary::WriteTreeSummary (ofstream& out)
{
  out << "\t" << xmlsum::TREESUM_START << endl;
  out << "\t\t" << xmlsum::NCOPY_START << " " << GetNCopies()
      << " " << xmlsum::NCOPY_END << endl;

  CacheSummaries(); //just in case.

  // If growth is on, we probably have to write out full intervals,
  // not just points and waits.
  // However, some forces may only come in a short form anyway, if they never
  // happened, or only happened once (like recombination, for example).
  // But IntervalData has this information stored in it now, so we're OK.

  string form = "short";  //Either 'short' or 'long'.
  DoubleVec1d sp; 
  DoubleVec1d sw;
  LongVec2d shortpick;

  vector<force_type> fvec;
  fvec.push_back(force_COAL);
  fvec.push_back(force_MIG);
  fvec.push_back(force_DISEASE);
  fvec.push_back(force_REC);

  bool somelong = false;

  for (unsigned long fvi=0; fvi < fvec.size(); fvi++) {
    Summary* currsum = GetSummary(fvec[fvi]);
    if (currsum) {
      if (!(currsum->GetShortness())) {
	form = "long";
	somelong = true;
      }
      sp = currsum->GetShortPoint();
      sw = currsum->GetShortWait();
      shortpick = currsum->GetShortPicks();

      out << "\t\t" << xmlsum::SHORTFORCE_START << " "
          << ToShortString(fvec[fvi])
	  << " " << form << endl;
      WriteShorts( out, sp, sw, shortpick);
      out << "\t\t" << xmlsum::SHORTFORCE_END << endl;
    }
    form = "short";
  }

  if (somelong) {
    out << "\t\t" << xmlsum::INTERVALS_START << endl;
    intervalData.WriteIntervalData(out);
    out << "\t\t" << xmlsum::INTERVALS_END << endl;
  }
  out << "\t" << xmlsum::TREESUM_END << endl;
}/* TreeSummary::WriteTreeSummary */

// writes a summary's shortwait and shortpoint to sumfile
void TreeSummary::WriteShorts (ofstream& out, DoubleVec1d& shortpoint, DoubleVec1d& shortwait, LongVec2d& shortpick) const
{
  vector<double>::iterator itstart;
  vector<double>::iterator itend;

  itstart = shortpoint.begin();
  itend	= shortpoint.end();
	
  out << "\t\t\t" << xmlsum::SHORTPOINT_START << " ";
  for(; itstart!= itend; ++itstart)
    out << *itstart << " ";
  out << xmlsum::SHORTPOINT_END << endl;
	
  itstart = shortwait.begin();
  itend	= shortwait.end();
	
  out << "\t\t\t" << xmlsum::SHORTWAIT_START << " ";
  for(; itstart!= itend; ++itstart)
    out << *itstart << " ";
  out << xmlsum::SHORTWAIT_END << endl;

  if (!shortpick.empty()) {
    LongVec2d::iterator fitstart = shortpick.begin();
    LongVec2d::iterator fitend = shortpick.end();
    out << "\t\t\t" << xmlsum::SHORTPICK_START << endl;
    for(; fitstart!= fitend; ++fitstart) {
      out << "\t\t\t\t" << xmlsum::SHORTPICK_FORCE_START << " ";
      LongVec1d::iterator istart = fitstart->begin();
      LongVec1d::iterator iend = fitstart->end();
      for(; istart!= iend; ++istart)
        out << *istart << " ";
      out << xmlsum::SHORTPICK_FORCE_END << endl;
    }
    out << "\t\t\t" << xmlsum::SHORTPICK_END << endl;
  }
}


// fills an individual tree summary with the appropriate summaries
// precondition:  last str read in was xmlsum::TREESUM_START
// postcondition: last str read in should be xmlsum::TREESUM_END
void TreeSummary::ReadInTreeSummary ( ifstream& in )
{
  string tag;
  in >> tag;
  
  SumFileHandler::ReadInCheckFileFormat("TreeSummary::ReadInTreeSummary",
				      xmlsum::NCOPY_START, tag);
  long int tt;
  in >> tt;
  nCopies = tt;
  in >> tag;
  SumFileHandler::ReadInCheckFileFormat("TreeSummary::ReadInTreeSummary",
				      xmlsum::NCOPY_END, tag);

  in >> tag;
  force_type type;
  string typeString;
  string form;
  while ( !in.eof() && tag != xmlsum::TREESUM_END ){
    if ( tag == xmlsum::SHORTFORCE_START ) {
      in >> typeString;
      type = ProduceForceTypeOrBarf(typeString);
      in >> form;
      Summary* current_sum = GetSummary(type);
      if (form=="short")
	current_sum->SetShortness(1);
      else
	current_sum->SetShortness(0);

      ReadInSummary(type, tag, in);
      in >> tag;
      SumFileHandler::ReadInCheckFileFormat("TreeSummary::ReadInTreeSummary",
					  xmlsum::SHORTFORCE_END, tag );
    }
    else if ( tag == xmlsum::INTERVALS_START ){
      ReadInIntervals(tag, in);
      SumFileHandler::ReadInCheckFileFormat("TreeSummary::ReadInTreeSummary",
					  xmlsum::INTERVALS_END, tag );
    }
    in >> tag;
  }
} /* TreeSummary::ReadInTreeSummary */

// fill a vector of doubles for one force (it doesn't know which one it does)
// precondition: next string read in should be xmlsum::SHORTPOINT_START
// postcondition: last string read in should be xmlsum::SHORTWAIT_END
// 		  next string to be read should be an end force tag
void TreeSummary::ReadInSummary ( force_type type, string& tag, ifstream& in )
{
  Summary* current_sum = GetSummary(type);

  in >> tag;
  SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInSummary", xmlsum::SHORTPOINT_START, tag );
  in >> tag;
  while ( tag != xmlsum::SHORTPOINT_END ) {
    current_sum->AddShortPoint( tag ); 
    in >> tag;
  }

  in >> tag;
  SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInSummary", xmlsum::SHORTWAIT_START, tag );
  in >> tag;
  while ( tag != xmlsum::SHORTWAIT_END ) {
    current_sum->AddShortWait( tag );
    in >> tag;
  }

  ForceSummary& fs = registry.GetForceSummary();
  if (fs.GetNLocalPartitionForces() > 0 && fs.CheckForce(force_REC)
      && type == force_COAL) {
    in >> tag;
    SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInSummary", xmlsum::SHORTPICK_START, tag );
    in >> tag;
    while ( tag != xmlsum::SHORTPICK_END ) {
      SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInSummary", xmlsum::SHORTPICK_FORCE_START, tag );
      in >> tag;
      StringVec1d newpicks;
      while ( tag != xmlsum::SHORTPICK_FORCE_END) {
        in >> tag;
        newpicks.push_back(tag);
      }
      current_sum->AddShortPicks( newpicks );
      in >> tag;
    }
  }

} /* TreeSummary::ReadInSummary */

void TreeSummary::ReadInIntervals( string& tag, ifstream& in )
{
  in >> tag;
  while (tag == xmlsum::FORCE_START) {

    string typeString;
    force_type type;
    in >> typeString;
    type = ProduceForceTypeOrBarf(typeString);
    in >> tag;
    SumFileHandler::ReadInCheckFileFormat("TreeSummary::ReadInIntervals",
					xmlsum::FORCE_END, tag );
    in >> tag;
    SumFileHandler::ReadInCheckFileFormat("TreeSummary::ReadInIntervals",
					xmlsum::ENDTIME_START, tag );
    in >> tag;
    double endtime = atof(tag.c_str());
    in >> tag;
    SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInInterval",
					 xmlsum::ENDTIME_END, tag );
    in >> tag;
    long oldstatus = 0;
    if (tag==xmlsum::OLDSTATUS_START) {
      in >> tag;
      oldstatus = atol(tag.c_str());
      in >> tag;
      SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInInterval",
					   xmlsum::OLDSTATUS_END, tag );
      in >> tag;
    }

    long newstatus = 0;
    if (tag==xmlsum::NEWSTATUS_START) {
      in >> tag;
      newstatus = atol(tag.c_str());
      in >> tag;
      SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInInterval",
					   xmlsum::NEWSTATUS_END, tag );
      in >> tag;
    }
 
    long activesites = 0;
    if (tag==xmlsum::ACTIVESITES_START) {
      in >> tag;
      activesites = atol(tag.c_str());
      in >> tag;
      SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInInterval",
					   xmlsum::ACTIVESITES_END, tag );
      in >> tag;
    }
 
    long recsite = 0;
    if (tag==xmlsum::RECSITE_START) {
      in >> tag;
      recsite = atol(tag.c_str());
      in >> tag;
      SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInInterval",
					   xmlsum::RECSITE_END, tag );
      in >> tag;
    }
 
    SumFileHandler::ReadInCheckFileFormat( "TreeSummary::ReadInInterval",
					 xmlsum::XPARTLINES_START, tag );
    in >> tag;
    LongVec1d xpartlines;
    while (tag != xmlsum::XPARTLINES_END) {
      xpartlines.push_back(atol(tag.c_str()));
      in >> tag;
    }

    in >> tag;
    LongVec2d partlines;
    
    if (tag==xmlsum::PARTLINES_START) {
      in >> tag;
      LongVec1d part_temp;
      while (tag != xmlsum::PARTLINES_END) {
	while (tag != ".") {
	  part_temp.push_back(atol(tag.c_str()));
	  in >> tag;
	}
	partlines.push_back(part_temp);
	part_temp.clear();
	in >> tag;
      }
      in >> tag;
    }

    LongVec1d partnerpicks;
    if (tag==xmlsum::PARTNERPICKS_START) {
      in >> tag;
      while (tag != xmlsum::PARTNERPICKS_END) {
	while (tag != ".") {
	  partnerpicks.push_back(atol(tag.c_str()));
	  in >> tag;
	}
	in >> tag;
      }
      in >> tag;
    }

    //The Summary function 'AddInterval' adds the interval to the interval
    // list while also setting up the appropriate pointers for that force
    // type's path through the intervals.
    Summary* sumptr = GetSummary(type);
    if (sumptr)
      sumptr->AddInterval(endtime, partlines, xpartlines, activesites,
			 oldstatus, newstatus, recsite, partnerpicks, type);
    else {
      //Assume the interval is free-floating, and add the interval directly.
      Interval* fakeint = NULL;
      intervalData.AddInterval(fakeint, endtime, partlines, xpartlines,
			       activesites, oldstatus, newstatus, recsite,
			       partnerpicks, type);
    }
    
  }
}
